/* Compile a St. Louis (Boss) style pointlist from segments. */

/*	(c) 1998, 2005, Michiel van der Vlist  */

/* This programme is copyright by Michiel van der Vlist. It is
	distributed as freeware for noncommercial use. Do not distribute
	modified versions of either the source or the object code.
	For commercial use, please contact the author */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <alloc.h>
#include <dir.h>
#include <dos.h>
#include <time.h>
#include <sys\timeb.h>
#include <fcntl.h>
#include <io.h>
#include <conio.h>
#include <bios.h>
#include <process.h>
#include <errno.h>

#define MYNAMEANDNUMBER "(c) 2005, Michiel van der Vlist, PA0MMV (2:280/5555)"
#define TRUE (1==1)
#define FALSE !TRUE
#define DEFAULTZONE 2
#define DEFAULTNET 280
#define DEFAULTREGION 28
#define MINENTRYS 1000
#define MAXENTRYS 10000
#define NAMELENG 37
#define MAXEXCLUDES 50
#define MAXINCLUDES 50
#define MAXUNPUBS 50
#define MAXREDUNTS 50
#define MAXCONVERTS 50
#define MAXBRATES 30
//#define VERSION "1.06 beta 2005.07.23"
#define VERSION  "1.06"
#define INPLISTN "P28-LIST"
#define INPLISTE ".LST"
#define FONEXT ".FON"
#define HEADEREXT ".H"
#define TRAILEREXT ".T"
#define STLOUISEXT ".STL"
#define UNKNOWN 0
#define STLOUIS_B 1
#define STLOUIS_P 2
#define FIDOUSER 3
#define FIDOUSER_R 4
#define NODELIST "REGION28"
#define LOGFILE "PLMAKE.LOG"
#define CONFIGFILE "PLMAKE.CFG"
#define DIFLIST "P28-DIFF"
#define MAXDAYS 8
#define MAXINPUTS 200
#define MAXMAGICS 5
#define DEFLISTNAME "Pointlist"
#define NOTUSER 0
#define USER 1
#define USERTYPE 0
#define REGULARTYPE 1
#define IPTYPE 2
#define EMAILTYPE 3
#define MAXFLAGLENGTH 65

typedef unsigned char boolean;
typedef unsigned char byte;
typedef char string[NAMELENG];
typedef char longstring[80];
typedef char verylongstring[255];
typedef char megalongstring[512];
typedef char hyperlongstring[1025];

typedef struct {unsigned int zone,net,node,point; } fidonumber;

typedef struct {  unsigned int zone,net,node,point;
						byte flags;
					}	p_entry;

/* bits in flags
	0 - duplicate number detected
	1 - duplicate name detected
	2 -
	3 - Boss not found in nodelist
	4-  Boss is on Hold
	5 - Boss is Down
	6 - Marked as deleted duplicate
	7 - Boss found in nodelist	*/

longstring pointlist,oldlist,nodelist,diflist,logfile,header,trailer,fonfile;
longstring inputlist[MAXINPUTS],inputfile,pcmergepath,magics[MAXMAGICS];
int inputtype[MAXINPUTS], kosheruserflagpos=-1;
verylongstring pstr, excludeflags, emailflags, ipflags, userflags;
hyperlongstring regularflags;
string excludes[MAXEXCLUDES], includes[MAXINCLUDES], listname=DEFLISTNAME;
string reduntlist[MAXREDUNTS], convertlist[MAXCONVERTS];
long int unpubs[MAXUNPUBS];
p_entry huge *p_list;
p_entry far *p_l1;
FILE *infile,*outfile,*dfile,*logf,*fonf;
unsigned long nrofrecords;
long int totnrpoints;
fidonumber exclnr[MAXEXCLUDES];
unsigned int crc, index[MAXENTRYS];
int an,nr_bosses,totnr_bosses,nrofskips=0;
int ext_charset=2, fridays=1;
int errornumber=0, prefpref=0, nrinputs=0, dayweek, nrredunts=0, nrconverts=0;
int nrexcludes=0, nrincludes=0, nrexnmbrs=0, nrunpubs=0, nrmagics=0;
unsigned int defaultzone=DEFAULTZONE, defaultnet=DEFAULTNET;
char pext[MAXEXT], brates[MAXBRATES][9], sep='|';
int killmask=8;
int trailingcomma=0, bosstrailcomma=0;
boolean errabort=FALSE, nobosses=FALSE, agecheck=TRUE;
boolean reverseflag=FALSE, logflag=FALSE, excludeorphan=FALSE, difflag=TRUE;
boolean nochecks=FALSE, no_output=FALSE, warnmissing=TRUE, skipcheck=FALSE;
boolean nodupnamecheck=FALSE, interactive=FALSE, insert=TRUE, mono=FALSE;
boolean allunpub=FALSE, fonfileopen=FALSE, singlewordnames=FALSE;
boolean diffonly=FALSE, Poss=FALSE;
int unpubbossphone=0, writefonfile=0;

#include "rmainder.h"

int getcommands(int argc, char *argv[]);
int scanoption(char *st);
int readconfig(char *s, int maxl);
int readflagcontrolfile(char *fn);
boolean substenvirons(char *s, int max);
int directory(char *match, char *fname);
void scandir(char *match);
void force_ext(char *buf, char *ext);
boolean wildcards(char *s);
boolean forbiddenext(char *f);
void findlist(char *fbuf, char *inlist, char *defname, char *defext,int ern,
				boolean check);
void timestamp(FILE *f,int s);
int Error(int er, long int ln, char *fn, char *nn, char *lin, int cpos,
							 char *nam, int cpos2, int namsiz, boolean errab);
void help(void);
void pstring(char *st,int n);
unsigned int chcrc(unsigned char ch, unsigned int crc);
unsigned int Writeline(char *b, FILE *f, unsigned int ch);
int scanlist(FILE *inf,char *infn);
void scanfidouserlist(FILE *inf, char *infn, int ltype);
void scanstlouislist(FILE *inf, char *infn);
void scanxlaxlist(FILE *inf, char *infn);
boolean isalphaq(char c);
char *Fgets(char *s,int n, FILE *fp);
void Strcpy(char huge *dest, char *source);
void STrcpy(hyperlongstring huge *dest, char *source);
void strcpY(char *dest, char huge *source);
int Strcmp(char huge *dest, char huge *source);
boolean issep(char c);
int cmp_nodenr(const void *elem1, const void *elem2);
int cmpnodenr(unsigned int *elem1, unsigned int *elem2);
void checknodelist(void);
int searchnodelist(char *rl);
void makediff(void);
int getchbeep(int maxtime);
int nextfriday(int *today, time_t *secsfri);
char lowasc(char c, boolean force);
int handlefunnychars(int lnumber, char *infn, char *lin,char *nam
										,int offs,int namsiz,boolean spaceok);
int handlehighascii(long int lnumber, char *infn, char *lin,char *nam
																,int offs,int namsiz);
char copyword(char *dest, char *source);
boolean isexclude(char *s);
boolean isinincludes(char *s);
boolean isexclnumber(unsigned int zo,unsigned int ne,unsigned int no,unsigned int po);
int editline(char *fn, char *line,char *text,int pos,int maxlen);
int removeletter(char *s, int pos);
int sort_function( const void *a, const void *b);
int cmp_file_age(char *f1, char *f2);
boolean isemptyline(char *s);
byte getflags(unsigned int z, unsigned int n, unsigned int f);
byte GETflagsL(unsigned int z, unsigned int n, unsigned int f);
byte GETflagsB(unsigned int z, unsigned int n, unsigned int f);
int extractname(char *nam, char *s, int l);
int stufname(char *nam, char *s, int pos, int l, int maxl);
boolean cleanupline(char *s,long int lnr,unsigned int zo,unsigned int ne,unsigned int no);
int copyfield(char *line, char *field, int siz, int pos);
int cleansysname(char *s,char *sys,long int lnr,int sp,int siz);
int cleanloc(char *s,char *loc,long int lnr,int lp,int siz);
int cleanname(char *s,char *name,long int lnr,int np,int siz);
boolean cleanphone(char *s,char *ph,long int lnr,int pp,boolean isboss,unsigned int zo,unsigned int ne,unsigned int no,long int bpnr);
void cleanbaud(char *s, char *b, long int lnr, int bp, boolean isunpub);
void cleanflags(char *s, char *f, long int lnr, int fp, char *sn);
void substunder(char *s);
void substspace(char *s);
void cnvrtphone(char *dest, char *scrc);
boolean iskosherflag(char *f, boolean user);
boolean isexcludeflag(char *f);
boolean isreduntflag(char *f, char *g);
char *isconvertflag(char *f);
int strncmpwild(char *s, char *d, int n);
void updatemagics(void);
int cmp_magics(const void *elem1, const void *elem2);
void substactfile(char *f, char *m, boolean mag);
int Findfirst(const char *pathname, struct ffblk *ff_blk,int attrib);
int Findnext(char *patname,struct ffblk *ff_blk);
boolean supermatch(char *s, char *d);
int getlastweek(int day, int *year);
boolean isleapyear(int year);
int whatyear(void);
boolean isvalidparam(char *s, int t);
boolean isip(char *s);
boolean isipquad(char *s);
boolean isipv6(char *s);
boolean isfqdn(char *s);
boolean isemailaddress(char *s);
boolean isoftype(char *s, int t);

/************************** Main ********************/

int main(int argc, char *argv[])
{
	int i,j,k,l,len,n,m,today,nextfri,next,next1,intype,res,year1,er;
	unsigned int znnr,ntnr,ndnr,pntnr,bznnr,bntnr,bndnr;
	long int nrwritten=0,nrskipped=0,linenr,lblinenr;
	unsigned int lbznnr=0,lbntnr=0,lbndnr=0,lastcrc=0;
	char buffer[256];
	char ffdrv[MAXDRIVE], ffpath[MAXPATH], ffname[MAXFILE], ffext[MAXEXT];
	const char *monthn[]={"***","Jan","Feb","Mar","Apr","May","Jun",
									"Jul","Aug","Sep","Oct","Nov","Dec"};
	const char *daysname[]={"Sunday","Monday","Thuesday","Wednesday",
										"Thursday","Friday","Saturday"};
	string version,nam;
	longstring fbuf,fbuf1,p1,p2,p3,curwd;
	verylongstring line;
	struct date datum;
	struct tm *timeptr;
	time_t secondsonfriday, secondsonlastfriday, start, et;
	fpos_t bpos,lastbosspos;
	unsigned long freememory;
	byte flgs;
	boolean lastbosshaspoints=TRUE, faultybossline,ok,skipline;

	time(&start);
	strcpy(version,VERSION);
	printf("\nPLMAKE %s, Pointlist make utility\n",version);
	printf(MYNAMEANDNUMBER);
	printf("\nNoncommercial use allowed.\n");

	pointlist[0]=0;
	nodelist[0]=0;
	diflist[0]=0;
	logfile[0]=0;
	pcmergepath[0]=0;
	header[0]=0;
	trailer[0]=0;
	for (i=0;i<MAXENTRYS;i++) index[i]=i;
	for (i=0;i<MAXBRATES;i++) brates[i][0]=0;
	if (argc==2 && (argv[1][0]=='?' || toupper(argv[1][0])=='H'))
	{
		help();
		exit(0);
	}

	if ((i=readconfig(buffer,sizeof(buffer)-1))!=0)
	{
		printf("Error reading config file, line %d.\n",abs(i),NULL,0);
		if (i<0) j=1; else j=0;
		Error(24+j,0,NULL,buffer,NULL,0,NULL,0,0,TRUE);
	}

	if (argc>1)
	{
		if ((i=getcommands(argc, argv))!=0)
		{
			if (i<0) j=1; else j=0;
			Error(24+j,0,NULL,argv[abs(i)],NULL,0,NULL,0,0,TRUE);
		}
	}
	if (logflag)
	{
		if (logflag==2) logf=fopen(logfile,"wt"); else logf=fopen(logfile,"at");
		if (logf==NULL) Error(4,0,logfile,logfile,NULL,0,NULL,0,0,TRUE);
		else
		{
			getdate(&datum);
			i=datum.da_mon; if (i<1 || i>12) i=0;
			fprintf(logf,"\n----------  %d-%s-%d, PLMAKE v %s\n"
						,datum.da_day,monthn[i],datum.da_year,version);
		}
	}

	/*********** alocate memory ***********/

	freememory=farcoreleft()-5000;
	nrofrecords=freememory/(sizeof(p_entry)+1);
	if (nrofrecords<MINENTRYS) Error(99,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
	if (nrofrecords>MAXENTRYS)
	{
		nrofrecords=MAXENTRYS;
		printf("\nPlenty of m");
	}
	else printf("\nM");
	printf("emory for about %u bosses.\n",nrofrecords);
	p_l1=farcalloc(nrofrecords,sizeof(p_entry));
	if (p_l1==NULL) Error(99,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
	p_list=p_l1;

	/*************  read the input lists, pass 1 *****************/
	nr_bosses=totnr_bosses=0;
	totnrpoints=0;

	if (nrinputs==0 && !diffonly) Error(16,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);

	for (i=0;i<nrinputs;i++)
	{
		findlist(fbuf,inputlist[i],"",INPLISTE,18,FALSE);
		infile=fopen(fbuf,"rt");
		if (infile==NULL) Error(15,0,inputlist[i],inputlist[i],
															NULL,0,NULL,0,0,TRUE);
		sprintf(pstr,"Pass1, scanning segment: %s ",fbuf);
		inputtype[i]=scanlist(infile,fbuf);
	}
	findlist(fbuf1,pointlist,INPLISTN,".???",0,FALSE);
	strcpy(pointlist,fbuf1);

	if (diffonly)
	{
		sprintf(pstr,"Diff debug mode. Only generation of diff file.\n");
		pstring(pstr,0);
		goto Makediff;
	}

/************ sort bosses by number **************/

	sprintf(pstr,"Sorting bosses..\n");
	pstring(pstr,0);

	qsort(index,nr_bosses,sizeof(unsigned int),cmp_nodenr);

//	for (i=0;i<nr_bosses;i++)
//	printf("%u:%u/%u.%u %02X %d %d\n",p_list[index[i]].zone,
//				p_list[index[i]].net,p_list[index[i]].node,
//				p_list[index[i]].point,p_list[index[i]].flags,i,index[i]);

//	checkdoublenumbers();

	if (nodelist[0]!=1 && !reverseflag && !nochecks) checknodelist();
	else skipcheck=TRUE;

/******* read input again (pass 2 ) and write the output file *******/

	if (!no_output)
	{
		nextfri=nextfriday(&today, &secondsonfriday);
		if (fridays==2) next1=today; else next1=nextfri;
		if (fridays<0)
		{
			year1=whatyear();
			for (i=0;i>fridays;i--) next1=getlastweek(next1,&year1);
		}
		fnsplit(pointlist,ffdrv,ffpath,ffname,ffext);
		if (strcmp(ffext,".???")==0)
		{
			sprintf(ffext,".%03d",next1);
			fnmerge(pointlist,ffdrv,ffpath,ffname,ffext);
		}
		strcpy(pext,ffext);
		next=0;
		if (isdigit(pext[1]) && isdigit(pext[2]) && isdigit(pext[3]))
															 sscanf(&pext[1],"%d",&next);
		else fridays=0;
		if (fridays==3) next1=next+1;
		if (fridays) sprintf(pext,".%03d",next1);

		if (fridays==0 || next1==next)
		{
			fnmerge(fbuf1,ffdrv,ffpath,ffname,".BAK");
			strcpy(oldlist,fbuf1);
			if ((infile=fopen(oldlist,"rt"))!=NULL)
			{
				fclose(infile);
				if (remove(oldlist)==-1)
					 Error(11,0,oldlist,oldlist,NULL,0,NULL,0,0,TRUE);
			}
			if ((infile=fopen(pointlist,"rt"))!=NULL)
			{
				fclose(infile);
				if (rename(pointlist,oldlist)==-1)
							 Error(13,0,pointlist,pointlist,NULL,0,NULL,0,0,TRUE);
				sprintf(pstr,"%s renamed %s\n",pointlist,oldlist);
				pstring(pstr,0);
			}
		}
		fnmerge(fbuf1,ffdrv,ffpath,ffname,".???");
		strcpy(oldlist,fbuf1);
		findlist(fbuf,oldlist,INPLISTN,".???",0,FALSE);
		if (fridays!=0)
		{
			fnsplit(pointlist,ffdrv,ffpath,ffname,ffext);
			fnmerge(pointlist,ffdrv,ffpath,ffname,pext);
		}
		outfile=fopen(pointlist,"w+t");
		if (outfile==NULL)
							Error(10,0,pointlist,pointlist,NULL,0,NULL,0,0,TRUE);
		strcpy(fonfile,pointlist);
		force_ext(fonfile,FONEXT);
		if (writefonfile>0)
		{
			fonf=fopen(fonfile,"wt");
			if (fonf==NULL)
						Error(10,0,fonfile,fonfile,NULL,0,NULL,0,0,TRUE);
			else fonfileopen=TRUE;
		}
		sprintf(pstr,"Writing output to %s\n",pointlist);
		pstring(pstr,0);
		crc=0;
		getdate(&datum);
		i=datum.da_mon; if (i<1 || i>12) i=0;
		if (fridays==1 || fridays<0)
		{
			if ( fridays==1) secondsonlastfriday=secondsonfriday;
			else secondsonlastfriday=secondsonfriday+(604800L*fridays);
			timeptr=localtime(&secondsonlastfriday);
			j=fprintf(outfile,";A %s for %s %d-%s-%d -- Day number %s : ",
					listname,daysname[timeptr->tm_wday],timeptr->tm_mday,
					monthn[timeptr->tm_mon+1],timeptr->tm_year+1900,&pext[1]);
		}
		else
		{
			fprintf(outfile,";A %s for %s %d-%s-%d -- Day number %s : ",
				listname,daysname[dayweek],datum.da_day,
				monthn[i],datum.da_year,&pext[1]);
		}
		fgetpos(outfile,&bpos);      /* remember position of checksum */
		fprintf(outfile,"00000\n");
		sprintf(buffer,";S Created by PLMAKE %s\n",version);
		crc=Writeline(buffer,outfile,crc);
		fnsplit(nodelist,ffdrv,ffpath,ffname,ffext);
		if (!skipcheck)
		{
			sprintf(buffer,";S Checked against nodelist : %s%s\n",ffname,ffext);
			crc=Writeline(buffer,outfile,crc);
		}
		strcpy(buffer,";S\n");
		crc=Writeline(buffer,outfile,crc);

		if (header[0]!=1)
		{
			if (header[0]==0)
			{
				strcpy(header,pointlist);
				force_ext(header,HEADEREXT);
			}
			infile=fopen(header,"rt");
			if (infile==NULL) Error(9,0,header,header,NULL,0,NULL,0,0,FALSE);
			else
			{
				sprintf(pstr,"Reading header file %s\n",header);
				pstring(pstr,0);
				while (Fgets(pstr,sizeof(pstr)-1,infile)!=NULL)
				{
					if (pstr[0]!=';')
					{
						strcpy(buffer,";S");
						if (pstr[0]) { strcat(buffer," "); strcat(buffer,pstr); }
						strcpy(pstr,buffer);
					}
					if (pstr[strlen(pstr)-1]==0x1B) pstr[strlen(pstr)-1]=0;
					if (pstr[strlen(pstr)-1]=='\n') pstr[strlen(pstr)-1]=0;
					strcat(pstr,"\n");
					crc=Writeline(pstr,outfile,crc);
				}
				fclose(infile);
			}
		}

		for (i=0;i<nrinputs;i++)
		{
			bznnr=bntnr=bndnr=0;
			strcpy(fbuf1,inputlist[i]);
			strcpy(fbuf,fbuf1);
			intype=inputtype[i];
			if ((intype==FIDOUSER || intype==FIDOUSER_R) && pcmergepath[0]!=0x7F) /* (reversed) Fidouser ? */
			{
				force_ext(fbuf,STLOUISEXT);
				if (cmp_file_age(fbuf,fbuf1)<=0)
				{
					/* In order to avoid overflow of the arguments list we change
						the working directory, so we need not specify the entire
						path for the FidoUser and St.Louis lists */

					if ((getcwd(curwd,sizeof(curwd)))==NULL)
										Error(19,0,curwd,curwd,NULL,0,NULL,0,0,TRUE);
					fnsplit(fbuf1,ffdrv,ffpath,ffname,ffext);
					fnmerge(p1,ffdrv,ffpath,NULL,NULL);
					p1[strlen(p1)-1]=0; /* remove trailng backslash */
					fnmerge(p2,NULL,NULL,ffname,ffext);
					strcpy(p3,p2);
					force_ext(p3,STLOUISEXT);
					sprintf(pstr,"Chanching working directory to %s\n",p1);
					pstring(pstr,0);
					if ((chdir(p1)==-1))   /* working directory to segments */
										Error(19,0,p1,p1,NULL,0,NULL,0,0,TRUE);
					sprintf(pstr,"Calling PcMerge to convert %s to %s\n",p2,p3);
					pstring(pstr,0);
					strcpy(p1,"p="); strcat(p1,p2);
					strcpy(p2,"f="); strcat(p2,p3);
					strcpy(p3,"n="); strcat(p3,nodelist);
					if (logflag) fclose(logf);
					j=spawnl(P_WAIT,pcmergepath,"PCMERGE",p1,p2,p3,"l","m","-o","x=[boss]",NULL);
					res=chdir(curwd);	/* return to previous directory */
					if (logflag)
					{
						logf=fopen(logfile,"at");
						if (logf==NULL) Error(4,0,logfile,logfile,NULL,0,NULL,0,0,TRUE);
					}
					if (res==-1) Error(19,0,curwd,curwd,NULL,0,NULL,0,0,TRUE);
					if (j<0 || j>10) Error(96,0,NULL,NULL,NULL,0,NULL,0,j,TRUE);
					sprintf(pstr,"Returning to PlMake in directory %s...\n",curwd);
					pstring(pstr,0);
				}
				intype=STLOUIS_B;
			}

			infile=fopen(fbuf,"rt");
			if (infile==NULL) Error(15,0,fbuf,fbuf,NULL,0,NULL,0,0,TRUE);
			znnr=defaultzone; ntnr=defaultnet;
			linenr=lblinenr=0;
			sprintf(pstr,"Pass2, reading segment: %s \n",fbuf);
			pstring(pstr,0);
			while (Fgets(line,sizeof(line)-2,infile)!=NULL)
			{
				linenr++;
				if (line[0]==';')
				{
					if (line[1]!='E') continue;
					else
					{					/* Error lines /are/ copied to list */
						strcat(line,"\n");
						crc=Writeline(line,outfile,crc);
						continue;
					}
				}
				if (line[0]==' ') continue;
				if (line[0]=='\n') continue;
				if (line[0]==0) continue;
				if (intype==2)
				{
					if (strnicmp(line,"ZONE,",5)==0) {sscanf(&line[5],"%u",&znnr); continue;	}
					if (strnicmp(line,"HOST,",5)==0) {sscanf(&line[5],"%u",&ntnr); continue;	}
					if (strnicmp(line,"POINT,",6)==0) strcpy(buffer,&line[5]);
					else if (line[0]==',')
					{
						sscanf(&line[1],"%u",&ndnr);
						sprintf(buffer,"Boss,%u:%u/%u",znnr,ntnr,ndnr);
						if (bosstrailcomma==2) strcat(buffer,",");
						bznnr=znnr; bntnr=ntnr; bndnr=ndnr;
						lblinenr=linenr;
					}
				}
				else
				{
					strcpy(buffer,line);
					if (strnicmp(buffer,"BOSS,",5)==0)
					{
						do
						{
							ok=TRUE;
							if (sscanf(&buffer[5],"%u:%u/%u",&bznnr,&bntnr,&bndnr)!=3)
							{
								if (sscanf(&buffer[5],"%u/%u",&bznnr,&bntnr,&bndnr)!=2)
								{
									ok=FALSE;
									res=Error(29,linenr,fbuf,NULL,buffer,6,buffer,6,sizeof(buffer)-1-5,FALSE);
									faultybossline=TRUE;
									if (res<0 || !interactive) break;
								}
								else bznnr=defaultzone;
							}
							faultybossline=FALSE;
							lblinenr=linenr;
						}	while (!ok && interactive);
						if (faultybossline) continue;
					}
					else
					{
						if (faultybossline) continue;

						if (buffer[0]!=',')
						{
							j=0;
							while ((line[j]) && line[j]!=',') j++;
							strcpy(buffer,&line[j]);
							j=min(j,sizeof(nam)-1);
							strncpy(nam,line,j); nam[j]=0;
							/* The keyword "Point" is allowed in the first field
								of a 'Boss' type pointlist to allow for the idio-
								syncrasies of the Z2PK. Totally ignore it.
								The keyword "Pvt is allowed to take care of
								some R24 incompatibility problems. A warning is
								given and it is removed. Let's do the same for
								"Down", "Hold" and "Kenl".
								All other keywords are flagged as an error and
								the line is rejected. */
							if (strcmp(nam,"Point")!=0)
							{
								if (stricmp(nam,"HOLD")==0 || stricmp(nam,"DOWN")==0 ||
									stricmp(nam,"PVT")==0 ||
									stricmp(nam,"KENL")==0) er=46;
								else er=28;
								Error(er,linenr,NULL,nam,line,1,NULL,0,0,FALSE);
								if (er==28) continue;
							}
						}
					}
				}
				flgs=getflags(bznnr,bntnr,bndnr);
				if ((killmask & 0x0F) & (flgs>>3)) continue;
				if (buffer[0]==',')
				{
					totnrpoints++;
					do
					{
						res=sscanf(&buffer[1],"%u",&pntnr)-3;
						j=1;
						while (isdigit(buffer[j])) j++;
						if (res!=-2 || buffer[j]!=',')
						{
							res=(Error(5,linenr,fbuf,NULL,buffer,j+1,buffer,j+1,sizeof(buffer)-1,FALSE));
						}
					} while (res>=0 && interactive);
					if (res==-1) continue;

					if (isexclnumber(bznnr,bntnr,bndnr,pntnr))
					{
						sprintf(nam,"%u:%u/%u.%u",bznnr,bntnr,bndnr,pntnr);
						Error(27,linenr,NULL,nam,buffer,2,NULL,0,0,FALSE);
						continue;
					}
					n=extractname(nam,buffer,sizeof(nam)-1);
					k=0;
					while (nam[k]==' ' || nam[k]=='_') k++; /* skip lead spcs */
					if (nam[k]==0 || isexclude(&nam[k]))
					{
						Error(23,linenr,NULL,nam,buffer,n+1,NULL,0,0,FALSE);
						continue;
					}
					else
					{
						do
						{
							res=0;
							l=k;
							skipline=FALSE; ok=TRUE;
							len=strlen(nam)-1; m=0;  /* go hunt for word separations */
							for (;l<len;l++) if (nam[l]==' ' || nam[l]=='_') m++;
							if (m==0)
							{
								ok=FALSE;
								if (singlewordnames || isinincludes(nam))
								{                        /* warn single word name */
									res=Error(37,linenr,NULL,nam,buffer,n+len+1,nam,len+2,sizeof(nam),FALSE);
									skipline=(res<0) && interactive;
								}
								else
								{
									res=Error(38,linenr,NULL,nam,buffer,n+len+1,nam,len+2,sizeof(nam),FALSE);
									skipline=(res<0);   /* reject single word name */
								}
							}
							if (res>0) stufname(nam,buffer,n,len+1,sizeof(buffer));
						}	while (interactive && !ok && res>0);
						if (skipline) continue;
					}
				}
				if (strnicmp(buffer,"BOSS,",5)==0)
				{
					if (!lastbosshaspoints)
					{
						crc=lastcrc; 			    /* backup to skip */
						fsetpos(outfile,&lastbosspos);	/* last pointless boss */
						sprintf(nam,"%u:%u/%u",lbznnr,lbntnr,lbndnr);
						Error(21,lblinenr,NULL,nam,NULL,0,NULL,0,0,FALSE);
					}
					if	(isexclnumber(bznnr,bntnr,bndnr,0))	/* boss+points exlc? */
					{
						sprintf(nam,"%u:%u/%u",bznnr,bntnr,bndnr);
						Error(30,linenr,NULL,nam,buffer,5,NULL,0,0,FALSE);
						continue;
					}
					else
					{
						lastbosshaspoints=FALSE;
						lbznnr=bznnr; lbntnr=bntnr; lbndnr=bndnr;
						fgetpos(outfile,&lastbosspos); /* remember this position */
						lastcrc=crc; 	 /* and the current crc */
						sprintf(nam,";\n");
						crc=Writeline(nam,outfile,crc);
					}
				}
				else
				{
					if	(isexclnumber(bznnr,bntnr,bndnr,0)) continue;
					lastbosshaspoints=TRUE;
				}
				cleanupline(buffer,linenr,bznnr,bntnr,bndnr);
				strcat(buffer,"\n");
				crc=Writeline(buffer,outfile,crc);
				if (buffer[0]==',') nrwritten++;
			}
			fclose(infile);
		}
		sprintf(pstr,"%ld Points read from %d segments.\n",totnrpoints,nrinputs);
		pstring(pstr,0);

		if (trailer[0]!=1)
		{
			if (trailer[0]==0)
			{
				strcpy(trailer,pointlist);
				force_ext(trailer,TRAILEREXT);
			}
			infile=fopen(trailer,"rt");
			if (infile==NULL) Error(9,0,trailer,trailer,NULL,0,NULL,0,0,FALSE);
			else
			{
				sprintf(pstr,"Reading trailer file %s\n",trailer);
				pstring(pstr,0);
				while (Fgets(pstr,sizeof(pstr)-1,infile)!=NULL)
				{
					if (pstr[0]!=';')
					{
						strcpy(buffer,";S");
						if (pstr[0]) { strcat(buffer," "); strcat(buffer,pstr); }
						strcpy(pstr,buffer);
					}
					if (pstr[strlen(pstr)-1]==0x1B) pstr[strlen(pstr)-1]=0;
					if (pstr[strlen(pstr)-1]=='\n') pstr[strlen(pstr)-1]=0;
					strcat(pstr,"\n");
					crc=Writeline(pstr,outfile,crc);
				}
				fclose(infile);
			}
		}

		fputc(26,outfile);	/* terminating EOF(^z) */
		fsetpos(outfile,&bpos);
		fprintf(outfile,"%05u",crc);
		fclose(outfile);

		sprintf(pstr,"%ld Points written.\n",nrwritten);
		pstring(pstr,0);
		nrskipped=totnrpoints-nrwritten;
		if (nrskipped)
		{
			sprintf(pstr,"%ld Points skipped.\n",nrskipped);
			pstring(pstr,0);
		}
		sprintf(pstr,"Checksum : %05u\n",crc);
		pstring(pstr,0);
	}

Makediff:

	farfree(p_l1);
	if (difflag && !no_output) makediff();
	if (nrmagics>0 && !diffonly) updatemagics();
	time(&et);	et=et-start;
	sprintf(pstr,"PLMAKE completed, elapsed time: %ld:%02ld:%02ld\n",
							et/3600,(et/60)%60,et%60);
	pstring(pstr,0);
	return(0);
}

/* scan input list to determine type and collect boss numbers */

int scanlist(FILE *inf, char *infn)
{
	char buf[256];
	int listtype=UNKNOWN;
	int i,n=0;

	while (Fgets(buf,sizeof(buf)-1,inf)!=NULL && n<10)
	{
		i=0;
		if (buf[0]==';' || !buf[0]) continue;
		n++;
		if (strnicmp(buf,"BOSS,",5)==0 && isdigit(buf[5]))
										{ listtype=STLOUIS_B; break; }

		if	(strnicmp(buf,"HOST,",5)==0 || strnicmp(buf,"ZONE",5)==0 ||
															strnicmp(buf,"REGION",7)==0)
		{
			while (buf[i]!=',') i++;
			if (isdigit(buf[i+1])) { listtype=STLOUIS_P; break; }
		}
		i=0;
		if (isalpha(buf[0]))
		{
			while (isalnum(buf[i]) || buf[i]==' ') i++;
			if (buf[i++]!=',') continue;
			if (buf[i]!=' ' && !isalpha(buf[i]) && buf[i]!='\t') continue;
			i++;
			while (isalnum(buf[i])) i++;
			if (buf[i]!=' ' && buf[i++]!='\t') continue;
			while (buf[i]==' ' || buf[i]=='\t') i++;
			if (!isdigit(buf[i])) continue;
			listtype=FIDOUSER; break;
		}
		else
		{
			while (isdigit(i)) i++;
			if (buf[i]==',') continue;
			while (isgraph(buf[i])) i++;
			while (buf[i]==' ' || buf[i]=='\t') i++;
			if (!isalpha(buf[i])) continue;
			listtype=FIDOUSER_R; break;
		}
	}
	rewind(inf);
	switch(listtype)
	{
		case	STLOUIS_P:	scanxlaxlist(inf,infn); break;
		case	FIDOUSER_R:
		case	FIDOUSER:	scanfidouserlist(inf,infn,listtype); break;
		case	STLOUIS_B:
		case	UNKNOWN:
		default:	scanstlouislist(inf,infn); break;
	}
	return (listtype);
}

/* scan a FidoUser or swapped FidoUser input pointlist */

void scanfidouserlist(FILE *inf, char *infn, int ltype)
{
	char buf[256],buf1[256];
	long int a=0,linenumber=0;
	int b=0,c,i,res,swoffset=0;
	unsigned int zo,ne,no,po;
	boolean skipline, ok;

	strcat(pstr," (");
	if (ltype==FIDOUSER_R) strcat(pstr,"Rev ");
	strcat(pstr,"FidoUser)\n");
	pstring(pstr,0);
	c=nr_bosses;

	while (Fgets(buf1,sizeof(buf1)-1,inf)!=NULL && buf1[0]!=0x1A)
	{
		linenumber++;
		skipline=FALSE;
		if (buf1[0]==';' || buf1[0]=='\n' || buf1[0]==0 ||
					strncmp(buf1,"---",3)==0 || isemptyline(buf1))
																				continue;
		if (ltype==FIDOUSER_R) strcpy(buf,buf1);
		else
		{
			i=1;
			while(!isspace(buf1[i]) && !isspace(buf1[i+1]) && buf1[i+1]) i++;
			while(!isdigit(buf1[i]) && buf1[i]) i++;
			strcpy(buf,&buf1[i]);
			swoffset=i;
		}
		zo=ne=no=po=0;

		do
		{
			if (sscanf(buf,"%u:%u/%u.%u",&zo,&ne,&no,&po)!=4)
			{
				if (sscanf(buf,"%u/%u.%u",&ne,&no,&po)==3)
				{
					res=Error(0,linenumber,infn,NULL,buf1,i+1+swoffset,buf,1,sizeof(buf)-1-i,FALSE);
					if (res<0) { skipline=TRUE; continue; }
					if (res==0)	{ zo=defaultzone; ok=TRUE; }
					continue;
				}
				else
				{
					if (nobosses)
					{
						if (sscanf(buf,"%u:%u/%u",&zo,&ne,&no)==3) skipline=TRUE;
						else
						{
							if (sscanf(buf,"%u/%u",&ne,&no)==2) skipline=TRUE;
							else
							{
								res=Error(7,linenumber,infn,NULL,buf1,i+1-swoffset,buf,1,sizeof(buf)-1-i,FALSE);
								if (res<0) { skipline=TRUE; continue; }
							}
						}
					}
					else
					{
						res=Error(7,linenumber,infn,NULL,buf1,i+1-swoffset,buf,1,sizeof(buf)-1-i,FALSE);
						if (res<0) { skipline=TRUE; continue; }
					}
				}
			}
			else ok=TRUE;
		}	while (!ok && !skipline);

		if (skipline) { a++; continue; }

		p_list[nr_bosses].flags=0;
		p_list[nr_bosses].zone=zo;
		p_list[nr_bosses].net=ne;
		p_list[nr_bosses].node=no;
		p_list[nr_bosses].point=0;
		nr_bosses++;
		if (nr_bosses>=nrofrecords-1) Error(95,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
		a++;
	}
	fclose(inf);

	if (nr_bosses-c)
	{

		qsort(&index[c],nr_bosses-c,sizeof(unsigned int),cmp_nodenr);
		b=1;
		for (i=c+1;i<nr_bosses;i++)
		{
			if (cmpnodenr(&index[i-1],&index[i])) b++;
		}
	}
	sprintf(pstr,"%ld Points and %d Bosses scanned.\n",a,b);
	pstring(pstr,0);
	totnr_bosses=totnr_bosses+b;
}

/* scan a St Louis input list */

void scanstlouislist(FILE *inf, char *infn)
{
	char buf[256];
	int b=0, i, j;
	long int a=0, linenumber=0;
	unsigned int zo, ne, no, po;
	boolean skipline, ok, foundboss=FALSE;

	strcat(pstr," (St. Louis, Boss)\n");
	pstring(pstr,0);

	while (Fgets(buf,sizeof(buf)-1,inf)!=NULL && buf[0]!=0x1A)
	{
		linenumber++;
		skipline=FALSE;
		if (buf[0]==';' || isemptyline(buf)) continue;

		if (strnicmp(buf,"BOSS,",5)==0)
		{
			foundboss=TRUE;
			do
			{
				ok=TRUE;
				if (sscanf(&buf[5],"%u:%u/%u",&zo,&ne,&no)!=3)
				{
					zo=defaultzone;
					if (sscanf(&buf[5],"%u/%u",&ne,&no)!=2)
					{
						Error(29,linenumber,infn,NULL,buf,6,NULL,0,0,FALSE);
						foundboss=FALSE;
						skipline=TRUE;
					}
				}
			}  while (!ok);
			if (!skipline)
			{
				p_list[nr_bosses].flags=0;
				p_list[nr_bosses].zone=zo;
				p_list[nr_bosses].net=ne;
				p_list[nr_bosses].node=no;
				p_list[nr_bosses].point=0;
				nr_bosses++;
				if (nr_bosses>=nrofrecords-1) Error(95,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
				b++;
			}
		}
		else
		{
			if (!foundboss) continue;
			do
			{
				ok=TRUE;
				j=0;
				if (!isdigit(buf[0])) while (buf[j] && buf[j]!=',') j++;
				if (!isdigit(buf[j+1]) || buf[j]!=',')
				{
					ok=FALSE;
					if (buf[j]==',') i=1; else i=0;
					if (Error(7,linenumber,infn,NULL,buf,i+1,buf,i+1,sizeof(buf)-1,FALSE)<0)
					{ skipline=TRUE; continue; }
				}
				i=j+1;
				sscanf(&buf[i],"%u",&po);
				if (po==0)
				{
					if (Error(8,linenumber,infn,NULL,buf,i+1,buf,i+1,sizeof(buf)-1,FALSE)<0)
					{ skipline=TRUE; continue; }
				}
			}
			while (!skipline && !ok);

			if (skipline) { a++; continue; }
			a++;
		}
	}
	fclose(inf);
	sprintf(pstr,"%ld Points and %d Bosses scanned.\n",a,b);
	pstring(pstr,0);
	totnr_bosses=totnr_bosses+b;
}

/* read an xlax style list */

void scanxlaxlist(FILE *inf, char *infn)
{
	char buf[256];
	int b=0, i;
	long int a=0, linenumber=0;
	unsigned int zo=defaultzone,ne=defaultnet,no=0,po=0,lzo=0,lne=0,lno=0;
	boolean skipline, ok;

	strcat(pstr," (St. Louis, Point)\n");
	pstring(pstr,0);
	zo=defaultzone;

	while (Fgets(buf,sizeof(buf)-1,inf)!=NULL && buf[0]!=0x1A)
	{
		linenumber++;
		skipline=FALSE;
		if (buf[0]==';' || isemptyline(buf)) continue;

		if (strnicmp(buf,"ZONE,",5)==0)
		{
			sscanf(&buf[5],"%u",&zo);
			ne=zo;
			no=0;
			continue;
		}
		if (strnicmp(buf,"HOST,",5)==0)
		{
			sscanf(&buf[5],"%u",&ne);
			no=0;
			continue;
		}
		if (strnicmp(buf,"REGION,",7)==0)
		{
			sscanf(&buf[7],"%u",&ne);
			no=0;
			continue;
		}
		if (buf[0]==',' || strnicmp(buf,"HUB,",4)==0 ||
			strnicmp(buf,"HOLD,",5)==0 || strnicmp(buf,"DOWN,",5)==0
			|| strnicmp(buf,"KENL,",5)==0 || strnicmp(buf,"PVT,",4)==0 )
		{
			i=0;
			while(buf[i]!=',' && buf[i]) i++;
			i++;
			sscanf(&buf[i],"%u",&no);
			continue;
		}
		if (strnicmp(buf,"POINT,",6)==0)
		{
			do
			{
				ok=TRUE;
				i=6;
				if (!isdigit(buf[i]))
				{
					ok=FALSE;
					if (Error(7,linenumber,infn,NULL,buf,i+1,buf,i+1,sizeof(buf)-1,FALSE)<0)
					{ skipline=TRUE; continue; }
				}
				sscanf(&buf[i],"%u",&po);
				if (po==0)
				{
					if (Error(8,linenumber,infn,NULL,buf,i+1,buf,i+1,sizeof(buf)-1,FALSE)<0)
					{ skipline=TRUE; continue; }
				}
			} while (!skipline && !ok);

			if (skipline) { a++; continue; }

			if (zo!=lzo || ne!=lne || no!=lno)
			{
				p_list[nr_bosses].flags=0;
				p_list[nr_bosses].zone=lzo=zo;
				p_list[nr_bosses].net=lne=ne;
				p_list[nr_bosses].node=lno=no;
				p_list[nr_bosses].point=0;
				nr_bosses++;
				if (nr_bosses>=nrofrecords-1) Error(95,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
				b++;
			}
			a++;
		}
	}
	fclose(inf);
	sprintf(pstr,"%ld Points and %d Bosses scanned.\n",a,b);
	pstring(pstr,0);
	totnr_bosses=totnr_bosses+b;
}

/* Scan the name for suspicious characters and report them.
	Suspect is anything else then a letter a "-"',a "'"
	and a known separator or a dash immidiatly following a separator.
	"High ascii letters" are treated elsewhere  */

int handlefunnychars(int lnumber, char *infn, char *lin,char *nam
										,int offs,int namsiz,boolean spaceok)
{
	int j=0, res=0;
	char c, d, e[10];

	while ((c=d=nam[j])!=0)
	{
		if ((!spaceok && c==' ') || ((c&0x80) && (c=lowasc(c,TRUE))=='?') ||
					 (!isalpha(c) && !issep(c) && c!=',' && c!='-' && c!='\'') ||
					 (j && issep(nam[j-1]) && c=='-') || (j==0 && c=='-'))
		{
			sprintf(e," '%c'.",d);
			res=Error(3,lnumber,infn,e,lin,offs+j+1,nam,j+1,namsiz,FALSE);
			if (res<0) break;
		}
		j++;
	}
	return(res);
}

/* Scan the name for characters with bit 7 high. Report and optionally
	replace them by an 'equivalent' ASCII character. */

int handlehighascii(long int lnumber, char *infn, char *lin,char *nam
																,int offs,int namsiz)
{
	int j, res=0;
	char e[60];

	j=0;
	while (nam[j])
	{
		if (nam[j] & 0x80)
		{
			if (ext_charset>=2)
			{
				sprintf(e," '%c' replaced by '%c'.",nam[j],lowasc(nam[j],FALSE));
				nam[j]=lowasc(nam[j],FALSE);
			}
			else sprintf(e," '%c'.",nam[j]);

			if (Error(3,lnumber,infn,e,lin,offs+j+1,nam,j+1,namsiz,FALSE)<0)
				{ res=-1; break; }
		}
		j++;
		if (res<0) break;
	}
	if (!interactive) if (ext_charset<0) res=-1;
	return(res);
}


/* See if the string contains a comma */

boolean hascomma(char *n)
{
	return(strchr(n,',')!=NULL);
}

/* substitute low ascii for high ascii */

char lowasc(char c, boolean force)
{
	int i;
	const char hitabel[]={	'','C','','u','','e','','a','','a','','a',
									'','a','','c','','e','','e','','e','','i',
									'','i','','i','','A','','A','','E','','a',
									'','A','','o','','o','','o','','u','','y',
									'','O','','U','','c',
									'','f','','a','','i','','o','','u','','n',
									'','N',0,0};

	if (!force)
	{
		if (!(c&0x80) || ext_charset<2) return(c);
		if (ext_charset==2) return('?');
	}
	else if (!(c&0x80)) return(c);

	i=0;
	while (hitabel[i])
	{
		if (c==hitabel[i]) return(hitabel[i+1]);
		i+=2;
	}
	return('?');
}


/* Compare nodenumbers for sort */

int cmp_nodenr(const void *elem1, const void *elem2)
{
	return(cmpnodenr((unsigned int*)elem1,(unsigned int*)elem2));
}

int cmpnodenr(unsigned int *elem1, unsigned int *elem2)
{
	int r;
	p_entry p,q;

	p=p_list[*elem1];
	q=p_list[*elem2];

	r=p.zone - q.zone;
	if (r==0)
	{
		r=p.net - q.net;
		if (r==0)
		{
			r=p.node - q.node;
			if (r==0)
			{
				r=p.point - q.point;
			}
		}
	}
	if (r>0) r=1;
	if (r<0) r=-1;
	if (reverseflag) r=0-r;
	return(r);
}

/* see if char is alpha or single quote */

boolean isalphaq(char c)
{
	if (c=='\'') return(TRUE);
	return(isalpha(c));
}

/* Check if the name is in the twit list */

boolean isexclude(char *s)
{
	int i;
	string nnam;

	strcpy(nnam,s);

	for (i=0;i<nrexcludes;i++)
	{
		if (stricmp(nnam,excludes[i])==0) return(TRUE);
	}
	return(FALSE);
}

/* Check if the number is in the twit list */

boolean isexclnumber(unsigned int zo,unsigned int ne,unsigned int no,unsigned int po)
{
	int i;

	for (i=0;i<nrexnmbrs;i++)
	{
		if (exclnr[i].zone==zo && exclnr[i].net==ne &&
			exclnr[i].node==no && exclnr[i].point==po) return(TRUE);
	}
	return(FALSE);
}

/* check if the name is in the allowed list for single word names. */

boolean isinincludes(char *s)
{
	int i;
	string nnam;

	strcpy(nnam,s);

	for (i=0;i<nrincludes;i++)
	{
		if (stricmp(nnam,includes[i])==0) return(TRUE);
	}
	return(FALSE);
}

char copyword(char *dest, char *source)
{
	char c;
	int i=0;

	while (c=source[i],c && !issep(c))
	{
		dest[i]=c;
		i++;
	}
	dest[i]=0;
	return(c);
}

boolean issep(char c)
{
	return(c==' ' || c=='.' || c=='_');
}

void checknodelist(void)
{
	longstring fbuf;
	int i,j=0,nr_missing=0,nr_hold=0,nr_down=0,aant;
	boolean first=TRUE;
	byte flgs;

	findlist(fbuf,nodelist,NODELIST,".???",0,agecheck);
	strcpy(nodelist,fbuf);
	sprintf(pstr,"Checking for bossless points against %s\n",nodelist);
	pstring(pstr,0);
	aant=searchnodelist(nodelist);
	if (aant<0)
	{
		sprintf(pstr,"Can't find %s\n",nodelist);
		pstring(pstr,1);
		printf("Continue without bossless point check? (Y/N) ");
		i=toupper(getchbeep(300));
		printf("\n");
		if (i!='Y' && i!='J' && i!='\r')
		{
			Error(97,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
		}
		skipcheck=TRUE;
		return;
	}
	sprintf(pstr,"%d nodes and %d bosses read.\n",an,aant);
	pstring(pstr,0);
	for (i=0;i<nr_bosses;i++)
	{
		flgs=p_list[index[i]].flags;

		if (((flgs & 0x40)==0) && ((flgs & 0x80)==0 || (flgs & 0x30)))
		{
			if (first)
			{
				sprintf(pstr,"Orphaned points found:\n");
				pstring(pstr,1);
				sprintf(pstr,"Number              Boss status\n");
				pstring(pstr,1);
				first=FALSE;
			}
			else
			{
				if (cmpnodenr(&index[i-1],&index[i])==0) continue;
			}
			j=0;
			j=j+sprintf(pstr,"%u:%u/%u",p_list[index[i]].zone,
				p_list[index[i]].net,p_list[index[i]].node);
			if (j<20) for (;j<20;j++) strcat(pstr," ");
			if (flgs & (killmask<<3) & 0X3F) strcat	(pstr,"*");
			else strcat(pstr," ");
			if (flgs &0x30)
			{
				if (flgs & 0x10) {strcat(pstr,"Hold\n"); nr_hold++;}
				else {strcat(pstr,"Down\n"); nr_down++;}
			}  else {strcat(pstr,"Not found\n"); nr_missing++;}
			pstring(pstr,0);
		}
	}
	sprintf(pstr,"%d bosses not found, %d hold, %d down.\n",
													nr_missing,nr_hold,nr_down);
	pstring(pstr,0);
}

int searchnodelist(char *rl)
{
	char buffer[256],status[15];
	int nrbosses,i,m,n,z,p,nlhandle;
	unsigned int zonenr,netnr=0,nodenr=0;
	boolean zoneflag,hostflag;
	FILE *nfile;

	nlhandle=open(rl,O_RDONLY|O_DENYNONE);
	if (nlhandle==-1) return(-1);
	nfile=fdopen(nlhandle,"rt");
	if (nfile==NULL) return(-1);
	i=n=m=p=z=0;                           /* everything points to start */
	nrbosses=an=0;				            /* number of nodes/points prcsd */
	zonenr=defaultzone;
	while (TRUE)                           /* loop for all lines until EOF */
	{
		if (fgets(buffer,sizeof(buffer)-1,nfile)==NULL)	/* get a line from the infile */
		{                                   /* exit if EOF */
			fclose(nfile);
			for (i=0;i<nr_bosses;i++)
			{
				if ((p_list[index[i]].flags & 0x80)==0) p_list[index[i]].flags |=0x08;
			}
			return(nrbosses);
		}
		if (buffer[0]==';') continue;  /* test for and skip comment line */
		i=0;
		hostflag=zoneflag=FALSE;
		an++;		/* bump nodes processed */	/* first field of line to status */
		while (buffer[i]!=',' && buffer[i]) status[i]=buffer[i++];
		if (buffer[i] != ',') Error(14,0,rl,NULL,buffer,i,NULL,0,0,TRUE);
		status[i]=0;                        /* nulll terminator */
		i++;                                /* point to next field */
		if (stricmp(status,"POINT")==0) continue;
		if (stricmp(status,"ZONE")==0)
		{
			zoneflag=hostflag=TRUE;
			sscanf(&buffer[i],"%d",&zonenr);
			netnr=zonenr;
			nodenr=0;
			p=z=0;
			while (zonenr>p_list[index[z]].zone)		/* hunt for the zone */
			{
				z++;
				if (z>=nr_bosses) {z=-1; break;}	/* if no points found for zone */
			}
			m=n=z;
		}
		if (z<0) continue;	  /* if no points for this zone */
		if (stricmp(status,"REGION")==0 || stricmp(status,"HOST")==0 || zoneflag)
		{
			hostflag=TRUE;
			m=n=z;                              /* point to start of zone */
			p=0;
			if (!zoneflag) sscanf(&buffer[i],"%d",&netnr);
			nodenr=0;
			while (netnr>p_list[index[n]].net)			/* hunt for the net */
			{
				n++;
				if (n>=nr_bosses || zonenr!=p_list[index[n]].zone )
				{
					n=-1;
					break;								/* if no points found for net */
				}
			}
		}
		if (n<0) continue;							/* skip if no points in net */
		m=n;												/* point to current net */
		if (!hostflag)	sscanf(&buffer[i],"%d",&nodenr);
		while (nodenr>p_list[index[m]].node)			   /* hunt for any points */
		{
			m++;
			if (m>=nr_bosses || netnr!=p_list[index[m]].net || zonenr!=p_list[index[m]].zone )
			{
				m=-1;
				break;									/* if no points found for net */
			}
		}
		if (m<0) continue;
		p=0;

		while (nodenr==p_list[index[m+p]].node && netnr==p_list[index[m+p]].net
				&& zonenr==p_list[index[m+p]].zone && (m+p)<nr_bosses)
		{
			if ((p_list[index[m+p]].flags & 0x40)==0)		/* skip deleted entrys */
			{
				p_list[index[m+p]].flags |= 0x80; /* mark as boss found */
				if (stricmp(status,"HOLD")==0) p_list[index[m+p]].flags |=0x10;
				if (stricmp(status,"DOWN")==0 || stricmp(status,"KENL")==0)
															p_list[index[m+p]].flags |=0x20;
				if (p==0) nrbosses++;
			}
			p++;                                /* next point in array */
		}
	}
/*	return(-1); */
}

void makediff(void)
{
	int i,j,k,l,m,d,r,s,xy,year1;
	unsigned long mem;
	hyperlongstring huge *b1, huge *b2;
	hyperlongstring far *b_o, far *b_n;
	hyperlongstring a1,a2;
	char *x, *y, oldext[MAXEXT];
	unsigned long nrlines;
	fpos_t huge *c1;
	fpos_t huge *c2;
	fpos_t far *c_1;
	fpos_t far *c_2;
	fpos_t fp;
	boolean finito=FALSE, iscom1, iscom2, newateof=FALSE;

	fnsplit(pointlist,NULL,NULL,NULL,pext);
	if (!(isdigit(pext[1]) && isdigit(pext[2]) && isdigit(pext[3])))
	{
		sprintf(pstr,"%s has non-nummeric extension, can't make diff file.\n"
																						,pointlist);
		pstring(pstr,1);
		return;
	}
	if (diffonly) /* This is a kludge it is for debug only */
	{
		strcpy(oldlist,pointlist);
		sscanf(&pext[1],"%d",&i);
		year1=whatyear();
		j=getlastweek(i,&year1);			/* get the daynumber of last week's Friday */
		sprintf(a1,".%03d",j);
		force_ext(oldlist,a1);
	}
	fnsplit(oldlist,NULL,NULL,NULL,oldext);
	if (oldlist[0]==0 || !(isdigit(oldext[1]) && isdigit(oldext[2]) && isdigit(oldext[3])))
	{
		sprintf(pstr,"Can't find old list to make diff file, sorry.\n");
		pstring(pstr,1);
		return;
	}
	if (diflist[0]==0) strcpy(diflist,DIFLIST);
	d=fnsplit(diflist,NULL,NULL,NULL,NULL);
	if ((d & EXTENSION)==0) strcat (diflist,pext);
	infile=fopen(oldlist,"rt");		if (infile==NULL) strcpy(a1,oldlist);
	outfile=fopen(pointlist,"rt");	if (outfile==NULL) strcpy(a1,pointlist);
	dfile=fopen(diflist,"wt");			if (dfile==NULL) strcpy(a1,diflist);
	if (infile==NULL || outfile==NULL || dfile==NULL)
			{ Error(22,0,NULL,a1,NULL,0,NULL,0,0,TRUE); return; }

	mem=farcoreleft()-5000;
	nrlines=mem/(2*(sizeof(verylongstring)+sizeof(fpos_t)));
	if (nrlines<20L) Error(99,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
	b_o=farcalloc(nrlines,sizeof(verylongstring));
	b_n=farcalloc(nrlines,sizeof(verylongstring));
	c_1=farcalloc(nrlines,sizeof(fpos_t));
	c_2=farcalloc(nrlines,sizeof(fpos_t));
	b1=b_o;
	b2=b_n;
	c1=c_1;
	c2=c_2;
	if (b1==NULL || b2==NULL) Error(99,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);

	sprintf(pstr,"Creating difference file %s\n",diflist);
	pstring(pstr,0);
	sprintf(pstr,"from %s and %s\n",oldlist,pointlist);
	pstring(pstr,0);
	fgets(a1,sizeof(a1)-1,infile);
	fputs(a1,dfile);
	fputs("D1\nA1\n",dfile);
	fgets(a1,sizeof(a1)-1,outfile);
	fputs(a1,dfile);

	do															/* while (!finito) */
	{
		i=0;
		do                                        /* loop as long as they */
		{                                         /* are the same */
			fgetpos(infile,&fp);					/* remember pos'ns of lines */
			c1[0]=fp;
			fgetpos(outfile,&fp);
			c2[0]=fp;
			x=fgets(a1,sizeof(a1)-1,infile);
			y=fgets(a2,sizeof(a2)-1,outfile);
			if (x!=NULL) STrcpy(&b1[0],a1); else STrcpy(&b1[0],"\n");
			if (y!=NULL) STrcpy(&b2[0],a2); else STrcpy(&b2[0],"\n");
			r=Strcmp(b1[0],b2[0]);
			i++;
		}  while(r==0 && x!=NULL && y!=NULL);

		if	(x==NULL && y==NULL) xy=0;
		if (x==NULL && y!=NULL) xy=1;
		if (x!=NULL && y==NULL) xy=2;
		if (x!=NULL && y!=NULL) xy=3;

		if (xy==0)							/* if both are at EOF */
		{
			if (i>1) fprintf(dfile,"C%d\n",i-1); /* write last line */
			finito=TRUE;                    			/* and finito */
		}

		if (xy==1)          			/* Old file at EOF */
		{                                			/* new file has lines left */
			if (i>1) fprintf(dfile,"C%d\n",i-1);
			i=1;
			do
			{
				y=fgets(a1,sizeof(a1)-1,outfile);
				if (y!=NULL) STrcpy(&b2[i],a1);  /* read lines into mem */
				i++;										/* and count them */
				if (i>=nrlines) Error(98,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
			}	while (y!=NULL);
				fprintf(dfile,"A%d\n",i-1);
				for (j=0;j<i-1;j++) fprintf(dfile,"%s",&b2[j]);
				finito=TRUE;                        	/* and diff finished */
		}

		if (xy==2)                	/* old file not at EOF */
		{
			if (i>1) fprintf(dfile,"C%d\n",i-1);
			finito=TRUE;									/* diff is finished */
		}

		if (xy==3)					/* both not at EOF and different */
		{
			if (i>1) fprintf(dfile,"C%d\n",i-1);
			for (i=l=m=1;i<nrlines&&!newateof;i++,l++,m++)
			{
				fgetpos(infile,&fp);
				c1[i]=fp;
				fgetpos(outfile,&fp);
				c2[i]=fp;
				x=fgets(a1,sizeof(a1)-1,infile);
				y=fgets(a2,sizeof(a1)-1,outfile);
				if (x!=NULL) STrcpy(&b1[i],a1);
							else { STrcpy(&b1[i],"\n"); l--; l=max(l,1); }
				if (y!=NULL) STrcpy(&b2[i],a2);
							else { STrcpy(&b2[i],"\n"); newateof=TRUE; }
				j=i;

				/*	we need to make an exception to avoid getting a resync on
					two lines consisting of a ';' or ";S" only. There are just
					too many of them, so 'false' syncs occur too often. */

				iscom1=(Strcmp(b1[i],";\n")==0 || Strcmp(b1[i],";S\n")==0);
				iscom2=(Strcmp(b2[i],";\n")==0 || Strcmp(b2[i],";S\n")==0);
				do
				{
					r=Strcmp(b1[i],b2[j]) || iscom1;
					s=Strcmp(b1[j],b2[i]) || iscom2;
					j--;
				}	while (j>=0 && r && s);
				j++;
				if (r==0 || s==0) break;
			}
			if (i>=nrlines) Error(98,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);

			iscom1=(Strcmp(b1[i-1],";\n")==0);
			iscom2=(Strcmp(b2[i-1],";\n")==0);

			if	(r==0 && s==0)
			{
				/* we need an exception again for the ';' case */
				if (i>1 && iscom1 && iscom2) { i--; l--; m--; }
				fprintf(dfile,"D%d\n",l);
				fprintf(dfile,"A%d\n",m);
				for (k=0;k<m;k++) fprintf(dfile,"%s",&b2[k]);
				fp=c1[i];
				fsetpos(infile,&fp);
				fp=c2[i];
				fsetpos(outfile,&fp);
			}

			if (r==0 && s!=0)
			{
				fprintf(dfile,"D%d\n",i);
				fp=c1[i];
				fsetpos(infile,&fp);
				fp=c2[0];
				fsetpos(outfile,&fp);
			}
			if (s==0 && r!=0)
			{
				fprintf(dfile,"A%d\n",i);
				for (k=0;k<i;k++) fprintf(dfile,"%s",&b2[k]);
				fp=c1[0];
				fsetpos(infile,&fp);
				fp=c2[i];
				fsetpos(outfile,&fp);
			}
		}		/* both not at EOF and different ends here */
	}	while (!finito);

	farfree(b_o);
	farfree(b_n);
	farfree(c_1);
	farfree(c_2);
	fclose(infile);
	fclose(outfile);
	if (fonfileopen) fclose(fonf);
	fclose(dfile);
}

/* Get the day number of the day one week ago */

int getlastweek(int day, int *year)
{
	if (day>=8) return (day-7);
	*year=*year-1;
	if (isleapyear(*year)) return(day+366-7);
	return(day+365-7);
}

/* determine if a year is a leapyear */

boolean isleapyear(int year)
{
	if (year%4!=0) return FALSE;
	if ((year%400)==0) return(TRUE);
	if ((year%100)==0) return(FALSE);
	return(TRUE);
}

/* get the current year */

int whatyear(void)
{
	struct date d;

	getdate(&d);
	return(d.da_year);
}

unsigned int Writeline(char *b, FILE *f, unsigned int ch)
{
	int i=0;

	while (b[i] != 0)
	{
		if (b[i]=='\n') ch=chcrc('\r',ch);
		ch=chcrc(b[i],ch);
		i++;
	}
	fputs(b,f);
	return(ch);
}

/* find an input list */

void findlist(char *fbuf, char *inlist, char *defname, char *defext,int ern, boolean check)
{
	int i;
	char fffdrv[MAXDRIVE], fffpath[MAXPATH], fffname[MAXFILE], fffext[MAXEXT];
	struct ftime ftimep;
	struct date fdate;
	struct time ftime;
	time_t timer;
	long filetime,fileage,days;
	int fhandle;

	strcpy(fbuf,inlist);
	i=fnsplit(fbuf,fffdrv,fffpath,fffname,fffext);
	if ((i & FILENAME) == 0)
	{
		strcpy(fffname,defname);
		strcpy(fffext,defext);
	}
	else if ((i & EXTENSION) == 0) strcpy(fffext,".???");
	fnmerge(fbuf,fffdrv,fffpath,fffname,fffext);
	if (directory(fbuf,inlist)==-1)
	{
		if (ern) Error(ern,0,NULL,inlist,inlist,0,NULL,0,0,TRUE);
		else { inlist[0]=0; return; }
	}
	fnsplit(inlist,NULL,NULL,fffname,fffext);
	fnmerge(fbuf,fffdrv,fffpath,fffname,fffext);
	strcpy(inlist,fbuf);
	if (check)
	{
		fhandle=open(inlist,O_RDONLY);
		if (fhandle==-1) Error(14,0,NULL,inlist,NULL,0,NULL,0,0,TRUE);
		if (getftime(fhandle,&ftimep)==-1) {close(fhandle); Error(14,0,NULL,inlist,NULL,0,NULL,0,0,TRUE);}
		close(fhandle);
		fdate.da_year=ftimep.ft_year+1980;
		fdate.da_mon=ftimep.ft_month;
		fdate.da_day=ftimep.ft_day;
		ftime.ti_hour=ftimep.ft_hour;
		ftime.ti_min=ftimep.ft_min;
		ftime.ti_sec=ftimep.ft_tsec;
		ftime.ti_hund=0;
		filetime=dostounix(&fdate,&ftime);		/* filetime naar seconden */
		fileage=time(&timer)-filetime;
		days=(fileage+42200l)/(86400l);			/* seconden naar dagen */
		if (days>MAXDAYS)
		{
			sprintf(pstr,"Warning: %s is older than %d days.\n",inlist,MAXDAYS);
			pstring(pstr,1);
			printf("Continue? (Y/N) ");
			i=toupper(getchbeep(300));
			printf("\n");
			if (i!='Y' && i!='J' && i!='\r')
			{
				Error(97,0,NULL,NULL,NULL,0,NULL,0,0,TRUE);
			}
		}
	}
}

/* compare the date/time stamps of two files */
/* return 1 if f1 is newer than f2 */

int cmp_file_age(char *f1, char *f2)
{
	struct ftime ftimep;
	struct date fdate;
	struct time ftime;
	long filetime1,filetime2;
	int fhandle;

	fhandle=open(f1,O_RDONLY);
	if (fhandle==-1) return(-1);
	if (getftime(fhandle,&ftimep)==-1) {close(fhandle); Error(14,0,NULL,f1,NULL,0,NULL,0,0,TRUE);}
	close(fhandle);
	fdate.da_year=ftimep.ft_year+1980;
	fdate.da_mon=ftimep.ft_month;
	fdate.da_day=ftimep.ft_day;
	ftime.ti_hour=ftimep.ft_hour;
	ftime.ti_min=ftimep.ft_min;
	ftime.ti_sec=ftimep.ft_tsec;
	ftime.ti_hund=0;
	filetime1=dostounix(&fdate,&ftime);		/* filetime naar seconden */

	fhandle=open(f2,O_RDONLY);
	if (fhandle==-1) Error(14,0,NULL,f2,NULL,0,NULL,0,0,TRUE);
	if (getftime(fhandle,&ftimep)==-1) {close(fhandle); Error(14,0,NULL,f2,NULL,0,NULL,0,0,TRUE);}
	close(fhandle);
	fdate.da_year=ftimep.ft_year+1980;
	fdate.da_mon=ftimep.ft_month;
	fdate.da_day=ftimep.ft_day;
	ftime.ti_hour=ftimep.ft_hour;
	ftime.ti_min=ftimep.ft_min;
	ftime.ti_sec=ftimep.ft_tsec;
	ftime.ti_hund=0;
	filetime2=dostounix(&fdate,&ftime);		/* filetime naar seconden */

	if (filetime1<filetime2) return(-1);
	if (filetime1>filetime2) return(1);
	return(0);
}

/* force the extension of a file name */

void force_ext(char *buf, char *ext)
{
	char drive[MAXDRIVE];
	char dir[MAXDIR];
	char file[MAXFILE];

	fnsplit(buf,drive,dir,file,NULL);
	fnmerge(buf,drive,dir,file,ext);
}

int getchbeep(int maxtime)
{

	time_t secsnow, secs, secs1;

	secs=secs1=time(&secsnow)-15;
	while (!kbhit())
	{
		time(&secsnow);
		if ((secsnow-secs1)>maxtime) return(0);
		if (secsnow-secs>15)
		{
			sound(600);
			delay(300);
			if (kbhit()) break;
			sound(800);
			delay(300);
			if (kbhit()) break;
			sound(1000);
			delay(300);
			nosound();
			secs=secsnow;
		}
	}
	nosound();
	return(getche());
}

/* get the day number for coming friday */

int nextfriday(int *today, time_t *secsfri)
{
	int day,dayw,daym,daysmore,month;
	struct tm *timeptr;
	time_t secsnow;

	time(&secsnow);
	timeptr=localtime(&secsnow);
	daym=timeptr->tm_mday;
	month=timeptr->tm_mon;
	day=timeptr->tm_yday;			/* get the day of the year */
	*today=day+1;							/* return that as wel */
	dayw=timeptr->tm_wday;        /* get day of week sunday=0 */
	dayweek=dayw;
	if (month>11||daym<1||daym>31) {*secsfri=0;*today=0;return(0);} /* error */
	dayw=dayw+2;
	if (dayw>6) dayw=dayw-7;		/* convert to friday=0 */
	if (dayw==0) daysmore=0; else daysmore=7-dayw;	/* days to go to friday */
	*secsfri=secsnow+((time_t)daysmore*86400L);
	timeptr=localtime(secsfri);
	day=timeptr->tm_yday;
	return(day+1);
}

/* directory searching */
/* This one is called just before opening the input file.
	Hunt for all matching files and if the extension is nummeric
	select the latest one */

int directory(char *match, char *fname)
{
	struct ffblk infoblk;
	char far *dtasave;
	char namef[MAXFILE],extf[MAXEXT],extf1[MAXEXT];
	int i,j,ret=-1;
	unsigned ffdate0=0, ffdate1=0;

	dtasave = getdta();				/* Save het org. disk transfer adres. */
	if (Findfirst(match,&infoblk,0)==0)
	{
		strcpy(fname,infoblk.ff_name);
		fnsplit(fname,NULL,NULL,namef,extf);
		strcpy(extf1,extf);
		if (isdigit(extf[1]) && isdigit(extf[2]) && isdigit(extf[3]))
		{
			sscanf(&extf[1],"%d",&i);
			ffdate0=infoblk.ff_fdate;
		}
		else i=0;

		while (Findnext(match,&infoblk)==0)
		{
			fnsplit(infoblk.ff_name,NULL,NULL,namef,extf);
			if (forbiddenext(extf)) continue;
			if (forbiddenext(extf1)) strcpy(fname,infoblk.ff_name);
			strcpy(extf1,extf);
			if (isdigit(extf[1]) && isdigit(extf[2]) && isdigit(extf[3]))
			{
				sscanf(&extf[1],"%d",&j);
				ffdate1=infoblk.ff_fdate;
				if (ffdate1>ffdate0 || (ffdate1==ffdate0 && j>i))
				{											/* select the latest */
					strcpy(fname,infoblk.ff_name);
					i=j;
					ffdate0=ffdate1;
				}
			}
		}
		fnsplit(fname,NULL,NULL,NULL,extf);
		if (forbiddenext(extf)) ret=-1; else ret=0;
	}
	setdta(dtasave);              /* org disk transfer adres herstellen */
	return(ret);
}

/* scan a directory to expand an input segment containing wildcards */
/* This one is called when building the input list after reading
	an i=filespec line in the config file. If no extension is given
	in the mask, the list of files matching is build without regard
	to the extension. This will eventually result in selecting
	the latest from each file whose name matches */

void scandir(char *match)
{
	struct ffblk infoblk;
	char far *dtasave;
	char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE],extf[MAXEXT];
	char fname[MAXPATH], namef[MAXPATH];
	int i,nrinp;
	boolean dup,noext=TRUE;

	noext=!(EXTENSION & fnsplit(match,drive,dir,name,extf));
	if (noext) force_ext(match,".???");
	nrinp=nrinputs;

	dtasave = getdta();				/* Save het org. disk transfer adres. */
	if (Findfirst(match,&infoblk,0)==0)
	{
		strcpy(fname,infoblk.ff_name);
		if (noext) force_ext(fname,"");		/* clear the extension if not */
		strcpy(namef,drive); 	            /* present in matchlist */
		strcat(namef,dir);
		strcat(namef,fname);
		strcpy(inputlist[nrinputs++],namef);
		while ((nrinputs<MAXINPUTS) && (Findnext(match,&infoblk)==0))
		{
			strcpy(fname,infoblk.ff_name);
			strcpy(namef,drive);
			strcat(namef,dir);
			strcat(namef,fname);
			if (noext)
			{
				force_ext(namef,"");
				dup=FALSE;
				for (i=nrinp;i<nrinputs;i++)
				{
					/* see if previous match found */
					if (stricmp(namef,inputlist[i])==0) { dup=TRUE; break; }
				}
				if (!dup) strcpy(inputlist[nrinputs++],namef);
			}
			else strcpy(inputlist[nrinputs++],namef);
		}
	}
	setdta(dtasave);              /* org disk transfer adres herstellen */

	if (nrinputs-nrinp>1)			/* sort what we found if more than one */
	{
		qsort(&inputlist[nrinp],nrinputs-nrinp,sizeof(inputlist[0]),sort_function);
	}
}

int sort_function( const void *a, const void *b)
{
	return( stricmp(a,b) );
}

/* see if a name contains wildcards */

boolean wildcards(char *s)
{
	int i=0;

	while (s[i])
	{
		if (s[i]=='?' || s[i]=='#' || s[i]=='@' || s[i]=='*') return(TRUE);
		i++;
	}
	return(FALSE);
}

/* test if extension is forbidden
	forbidden extensions are 999,BAK,BAT,COM,EXE,FON,H,T,Ann,Jnn,Lnn,Rnn,Snn and Znn */

boolean forbiddenext(char *f)
{
	char c;

	if (f[0]!='.') return(TRUE);

	c=toupper(f[1]);

	if (	stricmp(f,".BAK")==0 ||  stricmp(f,".BAT")==0 ||
			stricmp(f,".EXE")==0 ||  stricmp(f,".COM")==0 ||
			stricmp(f,".FON")==0 ||  stricmp(f,".T")==0 ||
			stricmp(f,".H")==0 || stricmp(f,".999")==0 ||
			(isdigit(f[2]) && isdigit(f[3]) && (c=='A'||c=='J'||c=='L'|| c=='R' ||c=='S' || c=='Z')))
			return(TRUE);

	return(FALSE);
}

/* copy to a string on the far heap with a huge pointer */

void Strcpy(char huge *dest, char *source)
{
	int i=0;

	while (source[i] && i<NAMELENG-1) (dest)[i]=source[i++];
	(dest)[i]=0;
}

void STrcpy(hyperlongstring huge *dest, char *source)
{
	int i=0;

	while (source[i]) (*dest)[i]=source[i++];
	(*dest)[i]=0;
}

/* copy from a string on the far heap with a huge pointer */

void strcpY(char *dest, char huge *source)
{
	int i=0;

	while ((source)[i] && i<NAMELENG-1) dest[i]=(source)[i++];
	dest[i]=0;
}

/* compare two strings on the far heap */

int Strcmp(char huge *dest, char huge *source)

{
	char a,b;
	int i,r;

	i=0;

	do
	{
		a=*(dest+i);
		b=*(source+i);
		r=(a-b);
		i++;
	}	while (r==0 && a!=0 && b!=0);

	return(r);
}

/* same as fgets except that it advances pointer to EOL
	after coping as many chars as possible and it substitutes tabs
	by two spaces. */

char *Fgets(char *s,int n, FILE *fp)
{
	int cc;
	int ii,jj=0;

	for (ii=0;ii<n-2;ii++)
	{
		cc=fgetc(fp);
		if (cc=='\t')
		{
			s[jj++]=' '; s[jj++]=' '; ii++; continue;
		}
		if (cc=='\r') continue;
		if (cc=='\n' || cc==EOF) break;
		s[jj++]=cc;
	}
	s[jj]=0;
	while (cc!='\n' && cc!=EOF) cc=fgetc(fp);
	if (cc==EOF && jj==0) return(NULL);
	else return(s);
}

/* read the config file if present. Return 0 if all ok. If an error
	occurs the line number is retured if the option is unknown. If the
	substitution of an environment variable fails, the line number * -1
	is returned. */

int readconfig(char *s, int maxl)
{
	FILE *f;
	int i=0,j;

	if ((f=fopen(CONFIGFILE,"rt"))==NULL) return(0);	/* if no config */
	printf("Reading config file.\n");
	while (Fgets(s,maxl,f)!=NULL)
	{
		i++;
		if ((j=scanoption(s))!=0) { fclose(f); return(j*i); }
	}
	fclose(f);
	return(0);
}

/* scan the command line for options and filenames */

int getcommands(int argc, char *argv[])
{
	int i,j;

	for (i=1;i<argc;i++)
	{
		if ((j=scanoption(argv[i]))!=0) return(j*i);
	}
	return(0);
}

/* Scan an option. Returns 0 if all ok. Returns 1 if unknown option
	and -1 if environment variable substitution fails */

int scanoption(char *st)
{
	int i,j=0,m,n;
	char c,d;
	verylongstring s;
	char ph[15];
	boolean minus=FALSE, plus=FALSE, comma=FALSE;

	strncpy(s,st,sizeof(s)-1);
	s[sizeof(s)-1]=0;
	if (!substenvirons(s,sizeof(s)-1)) return(-1);

	j=0;
	if (s[0]==';' || s[0]=='\r' || s[0]=='\n') return(0);

	while ((c=toupper(s[j++]))!=0)
	{
		if (c=='/') continue;
		if (c=='-') { minus=TRUE; continue; }
		if (c=='+') { plus=TRUE; continue; }
		d=toupper(s[j]);
		if (d=='=' || d==':') d=toupper(s[++j]);
		switch (c)
		{
			case 'A' :	agecheck=!minus; break;
			case 'B' :  if (d)
							{
								if (minus) nobosses=TRUE;

								for (m=0;m<sizeof(brates)/sizeof(brates[0]);m++)
																				brates[m][0]=0;
								m=0;
								while (&s[j] && m<sizeof(brates)/sizeof(brates[0]))
								{
									n=0;
									while (s[j] && s[j]!=',')
									{
										if (n<sizeof(brates[0])-1)
										{
											brates[m][n]=s[j];
											n++;
										}
										j++;
									}
									brates[m][n]=0;
									m++;
									if (s[j]) j++;
								}
							}
							else nobosses=minus;
							break;
			case 'C'	:  comma=FALSE;
							if (d)
							{
								if (isdigit(s[j])) sscanf(&s[j],"%d",&trailingcomma);
								else
								{
									m=0;
									while (s[j+m]) {if (s[j+m]==',') {comma=TRUE; break;} m++;}
									if (comma)
									{
										if (nrconverts<MAXCONVERTS)
										{
											if (s[j]=='\\') j++;
											strncpy(convertlist[nrconverts],&s[j],sizeof(convertlist[0])-1);
											convertlist[nrconverts][sizeof(convertlist[0])-1]=0;
											nrconverts++;
										}
									}
									else
									{
										if (minus)
										{
											pcmergepath[0]=0x7F;
											pcmergepath[1]=0;
										}
										else
										{
											strcpy(pcmergepath,&s[j]);
											if (pcmergepath[strlen(pcmergepath)-1]!='\\')
																		strcat(pcmergepath,"\\");
											strcat(pcmergepath,"PCMERGE");
										}
									}
								}
							}
							break;
			case 'D' :	difflag=!minus || plus;
							diffonly=plus;
							if (d) strcpy(diflist,&s[j]);
							break;
			case 'E' :  if (d)
							{
								if (isdigit(s[j]))
								{
									sscanf(&s[j],"%d",&ext_charset);
									if (minus) ext_charset=0-ext_charset;
								}
								else
								{
									if (s[j]=='\\') j++;
									strncpy(emailflags,&s[j],sizeof(emailflags)-1);
								}
								break;
							}
			case 'F'	:  if (plus) writefonfile=1;
							if (minus) writefonfile=-1;
							if (d)
							{
								if (d=='@' && s[j+1])
								{
									readflagcontrolfile(&s[j+1]);
									break;
								}
								else
								{
									if (d=='-' && s[j+1])
									{
										strncpy(excludeflags,&s[j+1],sizeof(excludeflags)-1);
										break;
									}
									else
									{
										strncpy(regularflags,&s[j],sizeof(regularflags)-1);
										regularflags[sizeof(regularflags)-1]=0;
										m=1;
										while (regularflags[m])
										{
											if (regularflags[m-1]==',' && toupper(regularflags[m]=='U'))
											{
												regularflags[m-1]=0;
												if (regularflags[m+1]) m++;
												if (regularflags[m]==',') m++;
												if (regularflags[m])
													strncpy(userflags,&regularflags[m],sizeof(userflags)-1);
												break;
											}
											m++;
										}
									}
								}
							}
							break;
			case 'G' :  if (d) strncpy(userflags,&s[j],sizeof(userflags)-1);
			case 'H'	:  if (minus) header[0]=1; else
							if (d) strcpy(header,&s[j]); else header[0]=0;
							break;
			case 'I'	:  if (minus) nrinputs=0; else
							if (d && nrinputs<MAXINPUTS)
							{
								if (wildcards(&s[j]))
								{
									scandir(&s[j]);
								}
								else strcpy(inputlist[nrinputs++],&s[j]);
							}
							break;
			case 'J' :  if (d) strncpy(ipflags,&s[j],sizeof(ipflags)-1);
							break;
			case 'K'	:	if (minus) killmask=0; else
							if (d) sscanf(&s[j],"%d",&killmask);
							break;
			case 'L'	:  if (minus) logflag=2; else logflag=TRUE;
							if (d) strcpy(logfile,&s[j]);
							else strcpy(logfile,LOGFILE);
							break;
			case 'M'	:  if (d)
							{
								if (nrmagics<MAXMAGICS)
								{
									strncpy(magics[nrmagics],&s[j],sizeof(magics[0])-1);
									magics[nrmagics][sizeof(magics[0])]=0;
									nrmagics++;
								}
							}
							else
							{
								warnmissing=!minus;
								if (plus) nrmagics=0;
							}
							break;
			case 'N'	:  if (minus) { nodelist[0]=1; skipcheck=TRUE; }
							else
							{
								if (d) strcpy(nodelist,&s[j]);
								else strcpy(nodelist,NODELIST);
							}
							break;
			case 'P'	:  Poss=plus;
							if (d) strcpy(pointlist,&s[j]);
							/* nrinputs=0; */
							break;
			case 'Q'	:	interactive=!minus;
							mono=(d=='M');
							break;
			case 'R' :  if (minus) fridays=0;
							else if (d)
							{
								if (isdigit(s[j]) || (s[j]=='-' && isdigit(s[j+1])))
																sscanf(&s[j],"%d",&fridays);
								else
								{
									if (nrredunts<MAXREDUNTS)
									{
										if (s[j]=='\\') j++;
										strncpy(reduntlist[nrredunts],&s[j],sizeof(reduntlist[0])-1);
										reduntlist[nrredunts][sizeof(reduntlist[0])-1]=0;
										nrredunts++;
									}
								}
							}
							break;
			case 'S'	:  if (d)
							{
								if (isdigit(s[j])) sscanf(&s[j],"%d",&bosstrailcomma);
								else if (!(isalpha(d)||d==','||d==':'||d=='_')) sep=d;
							}
							break;
			case 'T'	:  if (minus) trailer[0]=1; else
							if (d) strcpy(trailer,&s[j]);	else trailer[0]=0;
							break;
			case 'U'	:	if (d && nrunpubs<MAXUNPUBS)
							{
								if (s[j])
								{
									if (stricmp(&s[j],"BOSS")==0)
									{
										unpubbossphone=1;
										if (minus) unpubbossphone=-1;
									}
									else
									{
										if (stricmp(&s[j],"ALL")==0)
										{
											allunpub=!minus;
										}
										else
										{
											cnvrtphone(ph,&s[j]);
											sscanf(ph,"%ld",&unpubs[nrunpubs]);
											nrunpubs++;
										}
									}
								}
							}
							break;
			case 'X'	:  if (d && strnicmp(&s[j],"SINGLE",6)==0)
							{
								singlewordnames=plus && !minus; break;
							}
							if	(isdigit(d) && nrexnmbrs<MAXEXCLUDES)
							{
								i=sscanf(&s[j],"%u:%u/%u.%u",&exclnr[nrexnmbrs].zone,
									&exclnr[nrexnmbrs].net,&exclnr[nrexnmbrs].node,
									&exclnr[nrexnmbrs].point);
								if (i==4) nrexnmbrs++;
								else
								{
									printf("Problem with pointnumber: %s\n",&s[j]);
									return(1);
								}
							}
							else
							{
								if (plus)
								{
									if (d && nrincludes<MAXINCLUDES)
									{
										if (d=='\\') j++;
										if (s[j])
										{
											strncpy(includes[nrincludes],&s[j],sizeof(includes[0])-1);
											includes[nrincludes][sizeof(includes[0])-1]=0;
											nrincludes++;
										}
									}
								}
								else
								{
									if (d && nrexcludes<MAXEXCLUDES)
									{
										if (d=='\\') j++;
										if (s[j])
										{
											strncpy(excludes[nrexcludes],&s[j],sizeof(excludes[0])-1);
											excludes[nrexcludes][sizeof(excludes[0])-1]=0;
											nrexcludes++;
										}
									}
								}
							}
							break;
			case 'Y'	:	if (d) strncpy(listname,&s[j],sizeof(listname)-1);
							substspace(listname);
							break;
			case 'Z'	:	if (d)
							{
								if (sscanf(&s[j],"%u,%u",&defaultzone,&defaultnet)==2 ||
									sscanf(&s[j],"%u:%u",&defaultzone,&defaultnet)==2) break;
								else {sscanf(&s[j],"%u",&defaultzone); defaultnet=DEFAULTNET;}
							}
							break;
			case '#'	:	nodupnamecheck=!minus;
							break;
			case '!'	:	nodupnamecheck=nochecks=!minus;
							break;
			case '*'	:	nochecks=FALSE;
							no_output=TRUE;
							break;
			default	:	return(1);
		}
		break;
	}
	return(0);
}

/* Read an ERRFLAGS style flag control file */

int readflagcontrolfile(char *fn)
{
	int i,j,l,n,maxl;
	char c;
	FILE *cfile;
	char *p;
	longstring s;
	string keyword,flag;
	const char KEY[][10]={"FLAGS","IFLAGS","EFLAGS","USER","REDUNDANT","CONVERT"};
	boolean nomatch, colon, lower;

	userflags[0]=0;

	if ((cfile=fopen(fn,"rt"))==NULL) Error(22,0,NULL,fn,NULL,0,NULL,0,0,TRUE);
	sprintf(pstr,"Reading flag control file %s\n",fn);
	pstring(pstr,0);

	while (Fgets(s,sizeof(s)-1,cfile)!=NULL)
	{
		j=0;
		if (s[0]==';') continue;

		while(s[j] && s[j]!=' ' && j<sizeof(keyword)-2) keyword[j]=s[j++];
		keyword[j]=0;
		nomatch=TRUE;
		for (n=0;n<sizeof(KEY)/sizeof(KEY[0]);n++)
		{
			if (stricmp(keyword,KEY[n])==0) { nomatch=FALSE; break; }
		}
		if (nomatch) continue;
		while (s[j] && s[j]!=' ') j++;  /* skip to next word */
		if (!s[j]) continue;            /* if nothing follows */
		j++;
		switch (n)
		{
			case 0 : p=regularflags; maxl=sizeof(regularflags); goto flagcase;
			case 1 : p=ipflags; maxl=sizeof(ipflags); goto flagcase;
			case 2 : p=emailflags; maxl= sizeof(emailflags); goto flagcase;
			case 3 : p=userflags; maxl=sizeof(userflags);
			flagcase:
						while (s[j]!=0)				/* USER */
						{
							i=0; colon=lower=FALSE;
							while ((c=s[j])!=' ' && c)
							{
								flag[i]=s[j];
								if (s[j]==':') colon=TRUE;
								if (islower(s[j])) lower=TRUE;
								i++; j++;
							}
							if (s[j]==' ') j++;
							flag[i]=0;
							/* skip flags containing a colon or lower case letter */
							if (colon || lower) continue;
							l=strlen(p);
							if(l+strlen(flag)+2<maxl)
							{
								if (l!=0) strcat(p,",");
								strcat(p,flag);
							}
						}
						break;
			case 4 :	if (nrredunts<MAXREDUNTS)	/* REDUNDANT */
						{
							strncpy(reduntlist[nrredunts],&s[j],sizeof(reduntlist[0])-1);
							reduntlist[nrredunts][sizeof(reduntlist[0])-1]=0;
							i=0;
							while ((c=reduntlist[nrredunts][i])!=0)
							{
								if (c==' ') reduntlist[nrredunts][i]=',';
								i++;
							}
							nrredunts++;
						}
						break;
			case 5 :	if (nrconverts<MAXCONVERTS)	/* CONVERT */
						{
							strncpy(convertlist[nrconverts],&s[j],sizeof(convertlist[0])-1);
							convertlist[nrconverts][sizeof(convertlist[0])-1]=0;
							i=0;
							while ((c=convertlist[nrconverts][i])!=0)
							{
								if (c==' ') convertlist[nrconverts][i]=',';
								i++;
							}
							nrconverts++;
						}
						break;
		}
	}
	fclose(cfile);
	return(0);
}

/* Search for and substitute %strings% by the corresponding environment
	variables. Returns TRUE if succesful. */

boolean substenvirons(char *s, int max)
{
	char buf1[30], buf[256], *c;
	int i,j,k;

	buf1[0]=0;
	i=0; k=0;

	while (TRUE)
	{
		while(s[i] && s[i]!='%') { buf[k]=s[i]; i++; k++; }

		if (s[i]==0)
		{
			buf[k]=0; j=0;
			while (j<max-1 && buf[j]) { s[j]=buf[j]; j++; }
			s[j]=0;
			break;
		}
		i++;
		buf1[0]=0; j=0;
		while (s[i] && s[i]!='%')
		{
			buf1[j]=toupper(s[i]);
			i++; j++;
		}
		buf1[j]=0;

		if (s[i]=='%') i++; else return(FALSE);

		if ((c=getenv(buf1))==NULL) return(FALSE);
		j=0;
		while (c[j] && k<max-2) { buf[k]=c[j]; k++; j++; }
	}
	return(TRUE);
}

void cnvrtphone(char *dest, char *scrc)
{
	int i,k,l=1;

	i=strlen(scrc)-1;
	dest[0]=0;

	while (i>=0 && l<10)
	{
		if (!isdigit(scrc[i])) {i--; continue;}
		for (k=l;k>0;k--) dest[k]=dest[k-1];		/* move one char right */
		dest[0]=scrc[i];
		l++;
		i--;
	}
}

/* timestamp to log file */

void timestamp(FILE *f, int s)
{
	const char ind[]={'#','!','?','$','*',':','+','-','%','~','='};
	struct time ftime;

	gettime(&ftime);
	fprintf(f,"%c %2d:%02d:%02d  ",ind[s],ftime.ti_hour,ftime.ti_min,ftime.ti_sec);
}

/* Error handler */

int Error(int er, long int ln, char *fn, char *nn, char *lin, int cpos,
								char *nam, int cpos2, int namsiz, boolean errab)

/*	The error handler carries the following parameter:
	int er		-	Number of the error
	int ln		-	The line number in the input file where the error occurs
	char *fn		-	Pointer to the name of the input file.
	char *nn		-	Pointer to a string that should be displayed when
						reporting the error.
	char *lin	-	Pointer to a string where the line in error is stored
	int cpos		-	The position of the character in error in the line
						zero if the '^' pointer should not be printed.
	char *nam	-	A pointer to the string that can be edited. Can be
						the same as lin, but needn't be so
	int cpos2	-	The position of the char in nam, where the error occurs
	int namsiz	-	The size of the array were the editstring is stored
	boolean errab	True when this error should abort PlMake

	In interactive mode the return value is:
	-1 in case the line is rejected.
	 0 in case the line is accepted as is by the operator.
	+1 in case the line is edited.

	In non interactive mode, the return value is -1.
*/

{
	int i, editresult=0;
	char line[255], buf[6];

	line[0]=0;
	if (ln>0) sprintf(line,"Line number %ld : ",ln);

	switch(er)
	{
		case 0	:	strcat(line,"Warning: default ZONE (");
						sprintf(buf,"%d",defaultzone);
						strcat(line,buf);
						strcat(line,") used.");
						break;
		case 1	:  strcat(line,"Warning: Comma expected.");
						break;
		case 2	:	strcat(line,"Warning: First name expected.");
						break;
		case 3	:	strcat(line,"Warning: Suspect character");
						break;
		case 4	:	strcat(line,"Error opening logfile.");break;
		case 5	:	strcat(line,"Error: comma required.");break;
		case 7	:	strcat(line,"Error scannning number."); break;
		case 8	:  strcat(line,"0 is not a legal pointnumber."); break;
		case 9	:	strcat(line,"Header/trailer file not found: ");break;
		case 10	:	strcat(line,"Can't open output file: "); break;
		case 11	:	strcat(line,"Can't delete old .BAK file: "); break;
		case 12	:	strcat(line,"Can't open nodelist segment: "); break;
		case 13	:	strcat(line,"Can't rename old file to .BAK: "); break;
		case 14	:	strcat(line,"Error reading nodelist segment: "); break;
		case 15	:	strcat(line,"Error opening input point list: "); break;
		case 16	:	strcat(line,"No input segment list provided.");break;
		case 17	:	strcat(line,"Can't find a nodelist segment for input: ");break;
		case 18  :  strcat(line,"Can't find pointlist input segment: ");break;
		case 19  :	strcat(line,"Error changing/retrieving directory: ");break;
		case 21	:	strcat(line,"Warning: Boss without points : "); break;
		case 22	:	strcat(line,"Error opening file: "); break;
		case 23	:	strcat(line,"Warning: ");
						if (nn[0]==0) strcat(line,"Empty ");
						strcat(line,"Name rejected : "); break;
		case 24	:	strcat(line,"Unknown/unvalid option : "); break;
		case 25	:	strcat(line,"Unknown environment variable : "); break;
		case 26	:	strcat(line,"Warning: Name is single word : "); break;
		case 27	:	strcat(line,"Warning: Nodenumber rejected : "); break;
		case 28	:	strcat(line,"Line rejected, keyword unvalid: "); break;
		case 29	:	strcat(line,"Faulty BOSS line, ");
						if (interactive) strcat(line,"no boss, no points.");
						else strcat(line,"all his points skipped.");
						break;
		case 30	:	strcat(line,"Boss in exclude list, all his ponts skipped : "); break;
		case 31	:	strcat(line,"Error in phone nr; forced Unpublished: "); break;
		case 32	:	strcat(line,"Phonenr excluded; forced to Unpublished: "); break;
		case 33	:	strcat(line,"Error in speed field; forced to "); break;
		case 35	:	strcat(line,"Phonenr same as boss; forced to Unpublished: "); break;
		case 36	:	strcat(line,"Warning: Leading space(s) removed: "); break;
		case 37	:	strcat(line,"Warning: ");
		case 38	:	strcat(line,"Sysop name is one word only: ");break;
		case 39	:	strcat(line,"Empty field replaced by '_'"); break;
		case 40	:	strcat(line,"Duplicate flag removed: "); break;
		case 41	:	strcat(line,"Redundant flag ");
						strcat(line,fn);
						strcat(line," removed due to ");
						break;
		case 42	:	strcat(line,"Illegal flag removed: "); break;
		case 43	:	strcat(line,"Illegal user flag removed: "); break;
		case 44	:	strcat(line,"Flag moved to regular flag field: "); break;
		case 45	:	strcat(line,"Erroneus flag ");
						strcat(line,fn);
						strcat(line," replaced by ");
						break;
		case 46	:	strcat(line,"Invalid keyword removed: "); break;
		case 47	:	strcat(line,"No associated fqdn/ip-quad, flag removed: ");
						break;
		case 48	:  strcat(line,"No associated e-mail address, flag removed: ");
						break;
		case 95	:	strcat(line,"Too many bosses.."); break;
		case 96	:  sprintf(buf,"Error %d",namsiz);
						strcat(line,buf);
						strcat(line," calling PcMerge."); break;
		case 97	:  strcat(line,"User abort or timeout."); break;
		case 98 	:	strcat(line,"Files too different, can't complete diff file.");
						break;
		case 99	:	strcat(line,"Not enough memory."); break;
		default	:	;
	}
	if (nn!=NULL) sprintf(pstr,"%s%s\n",line,nn);
	else sprintf(pstr,"%s\n",line);
	pstring(pstr,2);
	if (ln>0 && lin!=NULL)
	{
		sprintf(pstr,"%s\n",lin);
		pstring(pstr,2);
		if (cpos>0)
		{
			for (i=0;i<cpos-1;i++) pstr[i]=' ';
			pstr[i++]='^';
			pstr[i]=0;
			strcat(pstr,"\n");
			pstring(pstr,0);
		}
	}
	if (warnmissing && er>=16 && er<21)
	{
		printf("Press any key to abort.\n");
		i=getchbeep(100);
	}
	if (errabort || errab)
	{
		sprintf(pstr,"Pointlist make aborted.\n");
		pstring(pstr,2);
		if (logflag) fclose(logf);
		exit(er);
	}
	else
	{
		if (interactive && nam!=NULL && namsiz>0)
		{
			if (nn!=NULL) strcat(line,nn);
			editresult=editline(fn,nam,line,cpos2-1,namsiz);
			switch(editresult)
			{
				case -1	:	sprintf(pstr,"Line rejected by operator\n");
								nrofskips++;
								break;
				case 0	:  sprintf(pstr,"Line accepted by operator\n");
								break;
				case 1	:	sprintf(pstr,"Line edited: %s\n",nam);
								break;
			}
			pstring(pstr,0);
		}
		else
		{
			if (er==38 || ((er>3 || (er==3 && ext_charset<0)) && er!=9 && er<26))
			{
				sprintf(pstr,"Offending line skipped\n");
				pstring(pstr,2);
				if (er>errornumber) errornumber=er;
				nrofskips++;
				editresult=-1;
			}
		}
	}
	return(editresult);
}

/* string output to screen and to log */

void pstring(char *st,int n)
{
	printf("%s",st);
	if (logflag)
	{
		if (n>=0) timestamp(logf,n);
		fprintf(logf,"%s",st);
	}
}

/* print short helptext */

void help(void)
{
	printf("\n");
	printf("\nPLMAKE [-a] [-b] [b=brt[,brt]..] [c=0..2] [c=pcmergepth|<cflaglst>]");
	printf("\n       [[-|+]d[=dif]] [[-]e={0|1|2}|<emailflaglst]>");
	printf("\n       [[+|-]f[=[-]<flaglst>|@cntrlfile]] [g=<userflaglst>] [[-]h[=headerfile]]");
	printf("\n       [i=includefile] [j=<ipflaglst>] [k=killmask] [[-]l[=logfile]");
	printf("\n       [m-magic filespec compressiontype] [[-]n[=nodelst]] [+][p[=pointlst]]");
	printf("\n       [[-]q[=mono]] [r={-n|0|1|2|3}|<rflaglst>] [s=0..2] [s=ipv6sep]");
	printf("\n       [[-]t[=trailerfile]] [u={phonnr|BOSS|ALL}]");
	printf("\n       [[+]x=excludename|excludenumber|SINGLEwordnames]");
	printf("\n       [y=listname] [z=defaultzone] [*]\n");
}

/* check if a line is empty */

boolean isemptyline(char *s)
{

	int i=0;
	boolean r=TRUE;

	while (s[i])
	{
		if (!isspace(s[i])) {r=FALSE; break;}
		i++;
	}
	return(r);
}

/* get the flag byte for a node */

byte getflags(unsigned int z, unsigned int n, unsigned int f)
{
	/* Because this routine is called again for each point, we remember
		the last one and if the boss is the same, we return that value
		without searching the aray again. This saves a lot of time. */

	static unsigned int lastzone=0, lastnet=0, lastnode=0, lastflag=0;

	if (z==lastzone && n==lastnet && f==lastnode) return(lastflag);

	lastzone=z; lastnet=n; lastnode=f;
	return(lastflag=GETflagsL(z,n,f));
//	return(lastflag=GETflagsB(z,n,f));
}

/* commented out for now, object too big...

byte GETflagsB(unsigned int z, unsigned int n, unsigned int f)
{
	/* this is a binary search through the array of stored
		boss nodes. It should be faster than a lineair search... */

	int i,*itemptr;
	int keynr;
	byte flg;

	keynr=nr_bosses;
	index[keynr]=keynr;
	p_list[keynr].zone=z; p_list[keynr].net=n; p_list[keynr].node=n;
	p_list[keynr].point=0;

	itemptr=bsearch(&keynr,index,nr_bosses,sizeof(int),cmp_nodenr);

	if (itemptr==NULL) { flg=0; printf("NULL "); }
	else flg=p_list[index[*itemptr]].flags;

	i=*itemptr;
if (f==43) printf("%u:%u/%u %u:%u/%u.%u  %02x %d %d\n",z,n,f,p_list[index[i]].zone,
				p_list[index[i]].net,p_list[index[i]].node,
				p_list[index[i]].point,flg,i,index[i]);
	return(flg);
} */

byte GETflagsL(unsigned int z, unsigned int n, unsigned int f)
{
	/* this part is a lineair search through the array of stored
		boss nodes. Naturally it is slow if the array gets big. */

	int i;

	for (i=0;i<nr_bosses;i++)
	{
		if (z==p_list[index[i]].zone && n==p_list[index[i]].net &&
					f==p_list[index[i]].node) break;
	}
	return(p_list[index[i]].flags);
}

/* extract the name from a line */

int extractname(char *nam, char *s, int l)
{
	int i=0,j=0,n;

	while (j<=3)
	{
		while (s[i] && s[i]!=',') i++;
		j++;
		i++;
	}
	j=0;
	n=i;
	while(s[i] && s[i]!=',' && j<l-1)
	{
		if (s[i]==' ') nam[j]='_'; else nam[j]=s[i];
		i++; j++;
	}
	nam[j]=0;
	return(n);
}

/* stuf the revised name back into the line */

int stufname(char *nam, char *s, int pos, int l, int maxl)
{
	int b,d,i,n,len;

	b=strlen(s);
	len=strlen(nam);
	d=len-l;
	if (d>0)   	/* new name longer than old */
	{
		n=min(b+d,maxl-1);
		for(i=n;i>pos+d;i--) s[i]=s[i-d];
		s[n]=0;
		for (i=0;i<len;i++) s[pos+i]=nam[i];
	}
	else
	{
		for (i=0;i<len;i++) s[pos+i]=nam[i];
		if (d<0)  /* if same length that's it */
		{
			i=pos+len;
			while (s[i-d]) { s[i]=s[i-d]; i++; }
			s[i]=0;
		}
	}
	return(0);
}


/* clean up the line */
/* return TRUE if the line is in error and must be skipped */

boolean cleanupline(char *s,long int lnr,unsigned int zo,unsigned int ne,unsigned int no)
{
	int i,j,k=0,sp,lp,np,pp,bp,fp;
	char key[10],nmbr[30],sys[65],loc[65],name[37],phone[20],baud[10],flgs[128];
	char phon[15];
	boolean isboss, isunpub, linehastrailingcomma=FALSE;
	static long int bosspnr=0;

	key[0]=sys[0]=loc[0]=name[0]=phone[0]=baud[0]=flgs[0]=phon[0]=0;

	i=strlen(s)-1;
	while (i>=0 && s[i]==' ') s[i--]=0;		/* remove trailing spaces */
	if (s[strlen(s)-1]==',') linehastrailingcomma=TRUE;
	substunder(s);						/* replace remaining spaces by '_' */
	i=0;
	k=copyfield(s,key,sizeof(key),k);

	k=copyfield(s,nmbr,sizeof(nmbr),k);

	sp=k;
	k=copyfield(s,sys,sizeof(sys),k);

	lp=k;
	k=copyfield(s,loc,sizeof(loc),k);

	np=k;
	k=copyfield(s,name,sizeof(name),k);

	pp=k;
	k=copyfield(s,phone,sizeof(phone),k);

	bp=k;
	k=copyfield(s,baud,sizeof(baud),k);

	fp=k;
	strncpy(flgs,&s[k],sizeof(flgs)-1);
	flgs[sizeof(flgs)-1]=0;

	i=strlen(key);
	if (i>1)
	{
		key[0]=toupper(key[0]);
		for (j=1;j<i;j++) key[j]=tolower(key[j]);
	}

	isboss=(strcmp(key,"Boss")==0);
	if (isboss)
	{
		cnvrtphone(phon,phone);
		sscanf(phon,"%ld",&bosspnr);
	}
	if (isboss && sys[0]==0)
	{
		sprintf(s,"%s,%s",key,nmbr);
		if (bosstrailcomma>0)
		{
			if (bosstrailcomma==2 || linehastrailingcomma) strcat(s,",");
		}
	}
	else
	{
		cleansysname(s,sys,lnr,sp,sizeof(sys));
		cleanloc(s,loc,lnr,lp,sizeof(loc));
		cleanname(s,name,lnr,np,sizeof(name));
		isunpub=cleanphone(s,phone,lnr,pp,isboss,zo,ne,no,bosspnr);
		cleanbaud(s,baud,lnr,bp,isunpub);
		if (strlen(flgs)>0) cleanflags(s,flgs,lnr,fp,sys);

		if (Poss && key[0]==0) strcpy(key,"Point"); /* to make a "POSS" list */
		sprintf(s,"%s,%s,%s,%s,%s,%s,%s",key,nmbr,sys,loc,name,phone,baud);
		if (strlen(flgs)>0) { strcat(s,","); strcat(s,flgs); }
		else if (trailingcomma>0)
		{
			if (trailingcomma==2 || linehastrailingcomma) strcat(s,",");
		}
	}
	return(FALSE);
}

/* Copy a field from the line.
	Truncate a field if larger than the size of the assigned array.
	Returns the position of the next field. */

int copyfield(char *line, char *field, int siz, int pos)
{
	int i=0;

	while (line[pos]!=',' && line[pos] && i<siz-1) field[i++]=line[pos++];
	field[i]=0;
	while (line[pos] && line[pos]!=',') pos++;
	if (line[pos]==',') pos++;
	return(pos);
}

/* substitute underlines for all spaces in a string */

void substunder(char *s)
{
	int i=0;

	while (s[i])
	{
		if (s[i]==' ') s[i]='_';
		i++;
	}
}

/* substitute spaces for all underlines in a string. This
	is the reverse of the substunder routine */

void substspace(char *s)
{
	int i=0;

	while (s[i])
	{
		if (s[i]=='_') s[i]=' ';
		i++;
	}
}

/* clean up the system name */

int cleansysname(char *s,char *sys,long int lnr,int sp, int siz)
{
	int res;

	if (sys[0]==0)
	{
		Error(39,lnr,NULL,NULL,s,sp+1,NULL,0,0,FALSE);
		strcpy(sys,"_");
	}
	res=handlehighascii(lnr,NULL,s,sys,sp,siz-1);

	return(res);
}

/* clean up the location name */

int cleanloc(char *s,char *loc,long int lnr,int lp, int siz)
{
	int res;

	if (loc[0]==0)
	{
		Error(39,lnr,NULL,NULL,s,lp+1,NULL,0,0,FALSE);
		strcpy(loc,"_");
	}
	res=handlehighascii(lnr,NULL,s,loc,lp,siz-1);

	return(res);
}


/* clean up the sysop name */

int cleanname(char *s,char *name,long int lnr,int np, int siz)
{
	int i,res;

	i=strlen(name)-1;

	if (i!=0)
	{
		while (i>=0 && (name[i]==' ' || name[i]=='_')) name[i--]=0;	/* remove trailing spaces */
	}
	if ((name[0]==' ' || name[0]=='_') && i!=0)
	{
		Error(36,lnr,NULL,name,s,np+1,NULL,0,0,FALSE);
		while (name[0]==' '|| name[0]=='_')
		{
			removeletter(name,0);	/* remove any leading spaces */
		}
	}

	res=handlehighascii(lnr,NULL,s,name,np,siz-1);

	return(res);
}


/* clean up the phone number */

boolean cleanphone(char *s,char *ph,long int lnr,int pp,boolean isboss,unsigned int zo,unsigned int ne,unsigned int no,long int bpnr)
{
	int i=0;
	boolean ok=FALSE, unpub=TRUE;
	char c, phonnr[10];
	long int pnr;

	if (stricmp(ph,"-Unpublished-")==0)
	{
		strlwr(ph);    	/* force proper casing without reporting */
		ph[1]=toupper(ph[1]);
		return(unpub);
	}

	if (allunpub)
	{
		strcpy(ph,"-Unpublished-");
		return(unpub);
	}

	while (ph[i])
	{
		ok=TRUE;
		c=toupper(ph[i]);
		if (!(isdigit(c) || c=='-' || c=='*' || c=='#' || c=='A'
												|| c=='B' || c=='C' || c=='D'))
																	{	ok=FALSE; break; }
		i++;
	}
	if (!ok)
	{
		Error(31,lnr,NULL,ph,s,pp+i+1,NULL,0,0,FALSE);
		strcpy(ph,"-Unpublished-");
	}
	else
	{
		unpub=FALSE;
		if (!isboss)
		{
			cnvrtphone(phonnr,ph);
			sscanf(phonnr,"%ld",&pnr);
			if (unpubbossphone && bpnr==pnr)
			{
				if (unpubbossphone>0) Error(35,lnr,NULL,phonnr,s,pp+1,NULL,0,0,FALSE);
				strcpy(ph,"-Unpublished-");
				unpub=TRUE;
			}
			else
			{
				for (i=0;i<MAXUNPUBS;i++)
				{
					if (unpubs[i]==pnr)
					{
						Error(32,lnr,NULL,phonnr,s,pp+1,NULL,0,0,FALSE);
						strcpy(ph,"-Unpublished-");
						unpub=TRUE;
						break;
					}
				}
			}
			if (!unpub && writefonfile>=0)
			{
				if (!fonfileopen)
				{
					fonf=fopen(fonfile,"wt");
					if (fonf==NULL)
							Error(10,0,fonfile,fonfile,NULL,0,NULL,0,0,TRUE);
					else fonfileopen=TRUE;
				}

				fprintf(fonf,"%u:%u/%u.%s\n",zo,ne,no,&s[1]);
			}
		}
	}
	return(unpub);
}

/* clean up the baud rate */

void cleanbaud(char *s, char *b, long int lnr, int bp, boolean isunpub)
{
	int i,n=0;
	boolean found=FALSE;

	if (brates[0][0]==0)				/* no allowed speeds defined */
	{										/* anthing nummeric goes */
		found=TRUE;
		while (b[n]) { if (!isdigit(b[n])) { found=FALSE; break; } n++; }
	}
	else                          /* only defined speeds allowed */
	{										/* check against the list */
		for (i=0;i<sizeof(brates)/sizeof(brates[0]);i++)
		{
			if (brates[i][0]==0) break;
			if (strcmp(b,brates[i])==0) { found=TRUE; break; }
		}
	}
	if (!found)
	{
		if (isunpub || brates[1][0]==0) strcpy(b,brates[0]);
		else strcpy(b,brates[1]);
		if (b[0]==0) strcpy(b,"2400");
		Error(33,lnr,NULL,b,s,bp+1+n,NULL,0,0,FALSE);
	}
}

/* Clean up the flags.

	The strategie is as folows:
	0)	Split the flag list into it's components.
	1) Translate everything to upper case except the Tyz flags.
	2) Convert all flags listed in the convert list. Ignore
		user or regular status in this phase.
	3) Remove IP and e-mail flags for which no associated address
		can be located.
	4) Remove duplicate flags. In case a duplicate is found, remove
		the rightmost flag. This will result in removing the user
		flag if a flag is listed in both the regular and the user section.
	5) Remove flags implied by another flag, again ignoring user status.
	6) Move user flags to the regular flag section if undefined as
		a user flag but defined as a regular flag and remove all
		otherwise undefined flags.
	7)	Put it all together again. 		*/

/* Het lijkt er op dat hier nog ergens een foutje in zit. Bij het
	runnen van de Z2PNT loopt ie vast tenzij 2, 4 en 5 uitgeschakeld
	worden. */

void cleanflags(char *s, char *f, long int lnr, int fp, char *sn)
{
	char fl[20][MAXFLAGLENGTH], movedfl[10][MAXFLAGLENGTH];
	char params[MAXFLAGLENGTH], *parm, *conv;
	int fpos[20];
	int i,j,k=0,upos=-1,mvd=0, type;
	boolean user=NOTUSER;

	for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++) fl[i][0]=0;
	for (i=0;i<(sizeof(movedfl)/sizeof(movedfl[0]));i++) movedfl[i][0]=0;

	/* split the flag list in its components */

	for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
	{
		j=0; fpos[i]=k;
		if (f[k]=='U')
		{
			upos=i;
			if (f[k+1]!=',')
			{
				k++;
				if (f[k]==0) break;
				strcpy(fl[i],"U");
				continue;
			}
		}
		while (f[k]!=',' && f[k]) fl[i][j++]=f[k++];
		fl[i][j]=0;
		if (f[k]==0) break; else k++;
	}
	for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
	{								 /* translate all execpt Txy to upper */
		if (fl[i][0]==0) break;
		if ((fl[i][0]=toupper(fl[i][0]))!='T')
		{
			j=1;               /* only the flag itself, not the parameter */
			while (fl[i][j] && fl[i][j]!=':') {fl[i][j]=toupper(fl[i][j]); j++;}
		}
	}
	f[0]=0;

	/* First replace obsolete flags like ISNDNx by their new
		equivalents if a convert flag list is defined. */

	if (nrconverts)	/* only if a convert list defined */
	{
		for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
		{
			if (fl[i][0]==0) break;
			if (fl[i][0]==0x7F) continue;
			parm=NULL;
			if ((conv=isconvertflag(fl[i]))!=NULL)
			{
				j=0;
				while (fl[i][j])
				{
					if (fl[i][j]==':') {parm=&fl[i][j]; break; }
					j++;
				}
				strncpy(params,parm,sizeof(params)-1);
				Error(45,lnr,fl[i],conv,s,fp+fpos[i]+1,NULL,0,0,FALSE);
				strcpy(fl[i],conv);
				if (parm!=NULL) strcat(fl[i],params);
			}
		}
	}

	/* Now remove all Ip and e-mail flags for which no address can
		be located. */

	for (type=IPTYPE;type<=EMAILTYPE;type++)
	{
		if (!isvalidparam(sn,type)) /* if there is an address in the BBS*/
		{	   						/*	name field all flags of this type are ok. */
			for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
			{
				if (fl[i][0]==0) break;
				if (fl[i][0]==0x7F) continue;
				if (!isoftype(fl[i],type)) continue;
				j=0; parm=NULL;
				while (fl[i][j])
				{
					if (fl[i][j]==':') {parm=&fl[i][j+1]; break; }
					j++;
				}
				if (parm!=NULL)
				{
					strncpy(params,parm,sizeof(params)-1);
					/* if a valid address is seen in the parameter field of a
						flag, all following flags of same type are ok */

						if (isvalidparam(params,type)) break;
				}
				Error(47-IPTYPE+type,lnr,NULL,fl[i],s,fp+fpos[i]+1,NULL,0,0,FALSE);
				fl[i][0]=0x7F;		/* remove the erroneus flags */
			}
		}
	}

	/* Then remove duplicate flags irrespective of regular or user
		status. If a duplicate is found remove the rightmost, so that
		a user flag is always removed if duplicated in the regular
		section. */

	for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
	{
		if (fl[i][0]==0) break;
		if (fl[i][0]==0x7F) continue;
		for (j=i+1;j<(sizeof(fl)/sizeof(fl[0]));j++)
		{
			if (fl[j][0]==0) break;
			if (fl[j][0]==0x7F) continue;
			if (strcmp(fl[i],fl[j])==0)
			{
				Error(40,lnr,NULL,fl[j],s,fp+fpos[j]+1,NULL,0,0,FALSE);
				fl[j][0]=0x7F;
			}
		}
	}

	if (nrredunts)			/* only if redundant flag list defined */
	{                    /* go find and remove redundant flags */
		for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
		{
			if (fl[i][0]==0) break;
			if (fl[i][0]==0x7F) continue;
			for (j=0;j<(sizeof(fl)/sizeof(fl[0]));j++)
			{
				if (fl[j][0]==0) break;
				if (i==j || fl[j][0]==0x7F) continue;
				if (isreduntflag(fl[i],fl[j]))
				{
					Error(41,lnr,fl[j],fl[i],s,fp+fpos[i]+1,NULL,0,0,FALSE);
					fl[j][0]=0x7F;
				}
			}
		}
	}

	if (regularflags[0])		/* only if flags defined */
	{
		for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++)
		{
			if (fl[i][0]==0) break;
			if (i==upos) user=USER;
			if (fl[i][0]==0x7F || i==upos) continue;
			if (!iskosherflag(fl[i],user))
			{
				/* if bad user flag is defined as regular flag */
				if (user && iskosherflag(fl[i],NOTUSER) &&
												mvd<sizeof(movedfl)/sizeof(movedfl[0]))
				{
					Error(44,lnr,NULL,fl[i],s,fp+fpos[i]+1,NULL,0,0,FALSE);
					strcpy(movedfl[mvd++],fl[i]);
					fl[i][0]=0x7F;
				}
				else
				{
					Error(42+user,lnr,NULL,fl[i],s,fp+fpos[i]+1,NULL,0,0,FALSE);
					fl[i][0]=0x7F;		/* remove the erroneus flags */
				}
			}
		}
	}

	/* Now go put it all together again skipping deleted flags */

	j=0;
	for (i=0;i<mvd;i++)		/* first the moved user flags */
	{
		if (j!=0) strcat(f,",");
		strcat(f,movedfl[i]);
		j++;
	}
	for (i=0;i<(sizeof(fl)/sizeof(fl[0]));i++) /* then the rest */
	{
		if (fl[i][0]==0 || fl[i][0]==0x7F) continue;	/* skip deleted or empty flags */
		if (j!=0) strcat(f,",");		/* first non-empty flag? */
		strcat(f,fl[i]);					/* paste all together */
		j++;
	}
	/* check for a user flag field containing a single 'U' flag.
		If found, remove it. */

	j=strlen(f);
	if (j>=2 && f[j-1]=='U' && f[j-2]==',') f[j-2]=0;
}

/* Check if a string matches one of the flag list */

boolean iskosherflag(char *f, boolean user)
{
	int i,l;
	char dest[MAXFLAGLENGTH];

	l=0;
	while (f[l] && f[l]!=':') dest[l]=f[l++];
	dest[l]=0;

	if (isexcludeflag(dest)) return(FALSE);
	if (user)
	{
		 if (isoftype(dest,USERTYPE)) return(TRUE);
	}
	else
	{
		for (i=REGULARTYPE;i<=EMAILTYPE;i++) if (isoftype(dest,i)) return(TRUE);
	}
	return(FALSE);
}

/* see if a string is in the excluded flag list */

boolean isexcludeflag(char *f)
{
	int i,j=0,l;

	if  (!excludeflags[0]) return(FALSE);

	l=strlen(f);

	do
	{
		i=0;
		while (excludeflags[j+i] && excludeflags[j+i]!=',') i++;
		if (i==l && strncmpwild(&excludeflags[j],f,l)==0) return(TRUE);
		j=j+i+1;
	} while (excludeflags[j-1]);
	return(FALSE);
}

boolean isreduntflag(char *s, char *d)
{
	int i,j,k,l;
	char source[20], dest[20], ss[MAXFLAGLENGTH], dd[MAXFLAGLENGTH];

	if (s[0]==0x7F || d[0]==0x7F) return(FALSE);

	l=0;			/* consider ':' as end of flag */
	while (s[l] && s[l]!=':') ss[l]=s[l++];
	ss[l]=0;
	l=0;
	while (d[l] && d[l]!=':') dd[l]=d[l++];
	dd[l]=0;


	for (i=0;i<nrredunts;i++)
	{
		j=0;
		while (reduntlist[i][j]!=',' && reduntlist[i][j]) source[j]=reduntlist[i][j++];
		source[j]=0;
		if (strncmpwild(source,ss,sizeof(source))!=0) continue;
		if (reduntlist[i][j]!=',') continue;
		do
		{
			k=0;
			j++;
			while (reduntlist[i][j]!=',' && reduntlist[i][j]) dest[k++]=reduntlist[i][j++];
			dest[k]=0;
			if (strncmpwild(dest,dd,sizeof(dest))==0) return(TRUE);
		}  while (reduntlist[i][j]==',');
	}
	return(FALSE);
}

/* See if a flag is in the convert list, if so return a pointer
	to the string it should be replaced with, else return NULL */

char *isconvertflag(char *f)
{
	int i,j,l;
	char dest[MAXFLAGLENGTH];

	l=0;
	while (f[l] && f[l]!=':') dest[l]=f[l++];	/* ':' is end of flag */
	dest[l]=0;

	for (i=0;i<nrconverts;i++)
	{
		j=0;
		while (convertlist[i][j] && convertlist[i][j]!=',') j++;
		if (strncmp(dest,convertlist[i],max(j,l))==0)
		{
			if (convertlist[i][j]==',')	return(&convertlist[i][j+1]);
		}
	}
	return(NULL);
}

/* compare flag strings, respecting wild caracters in the source

	? matches one character.
	$ matches one character A-Z, a-z
	@ matches one character 0-9
	< matches the previous character, or any character if the first
	* matches all chars till end of string		 */

int strncmpwild(char *s, char *d, int n)
{
	int i;

	for (i=0;i<n;i++)
	{
		if (s[i]==0 && d[i]==0) break;
		switch (s[i])
		{
			case '<'	:	if (i) { if (d[i]!=d[i-1]) return(-1); break; }
			case '?'	:	if (!d[i]) return(-1);
							break;
			case '$'	:	if (!isalpha(d[i])) return(-1);
							break;
			case '@'	:	if (!isdigit(d[i])) return(-1);
							break;
			case '*'	:	return(0);
			default	:	if (s[i]!=d[i]) return(d[i]-s[i]);
							break;
		}
	}
	return(0);
}

/* see if a string qualifies as a valid first parameter parameter
	for this type of flag */

boolean isvalidparam(char *s, int t)
{
	switch(t)
	{
		case IPTYPE		:	return(isip(s));
		case EMAILTYPE	:	return(isemailaddress(s));
	}
	return(FALSE);
}


/* See if a string qualifies as an address for an IP flag */

boolean isip(char *s)
{
	return(isfqdn(s) || isipquad(s) || isipv6(s));
}

/* See if a string qualifies as an ip quad */

/* four nummeric values in the range 0-254,0-254,0-254,0-255
	separated by dots.
	The following ranges are reserved for private networks and
	hence invalid:
	192.168.0.1 through and up to 192.168.0.254
	10.0.0.1 through and up to 10.255.255.254
	172.16.0.1 through and up to 172.31.255.254

	Also the local host: 127.0.0.1 and 0.0.0.0 are  invalid */

boolean isipquad(char *s)
{
	int n1,n2,n3,n4,dots;
	long unsigned int ipadr;

	dots=sscanf(s,"%d.%d.%d.%d",&n1,&n2,&n3,&n4);
	if (dots<4) return (FALSE);			/* too few fields */

	if (n1<0 || n2<0 || n3<0 || n4<0) return (FALSE);
	if (n1>254 || n2>254 || n3>254 || n4>255) return (FALSE);

	ipadr=(((unsigned long)n1)<<24)+(((unsigned long)n2)<<16)+
										(((unsigned long)n3)<<8)+(unsigned long)n4;
	if (ipadr==0) return(FALSE);
	if (ipadr>=(192L<<24)+(168L<<16)+1L && ipadr<=(192L<<24)+(168L<<16)+254L)
																				return(FALSE);
	if (ipadr>=(10L<<24)+1 && ipadr<=(10L<<24)+(255L<<16)+(255L<<8)+254L)
																				return(FALSE);
	if (ipadr>=(172L<<24)+(16L<<16)+1L && ipadr<=(172L<<24)+(31L<<16)+(255L<<8)+254L)
																				return(FALSE);
	if (ipadr==(127L<<24)+1L) return(FALSE);
	return (TRUE);
}

/* See if a string qualifies as an ipv6 address */

/*	The separator character (colon in the IP world) is defined in the
	global variable sep, default also a colon in this test program.

	Eight hexadecimal values in the range 0-FFFF separated by a SEP char.
	One or more successive zero values may once be represented by
	two successive separators. F.e. 1234::6789:abcd expands to
	1234:0:0:0:0:0:6789:abcd.

	All zeros is invalid. */

boolean isipv6(char *s)
{
	int i=0,n1,n2,n3,n4,n5,n6,n7,n8,seps=0,nrseps=0,doubleseppos=0;
	char c,t[40],tt[40];

/*	first substitute the separator character by a comma, count them and
	note the position of two successive separators on the fly */

	while (s[i] && i<sizeof(t)-2)
	{
		c=s[i];
		if (c==sep)
		{
			c=',';
			nrseps++;
			if (s[i+1]==sep)
			{
				if (!doubleseppos) doubleseppos=i; else return FALSE;
			}
		}
		t[i++]=c;
	}
	t[i]=0;

/* Now expand the string if there is a ddouble separator present */

	if (doubleseppos)
	{
		strcpy(tt,&t[doubleseppos+1]);
		t[doubleseppos]=0;
		for (i=0;i<8-nrseps;i++) strcat(t,",0");
		strcat(t,tt);
	}
	seps=sscanf(t,"%x,%x,%x,%x,%x,%x,%x,%x",&n1,&n2,&n3,&n4,&n5,&n6,&n7,&n8);
	if (seps<8) return (FALSE);			/* too few fields */

	if ((n1+n2+n3+n4+n5+n6+n7+n8)==0) return (FALSE);

	return (TRUE);
}

/* see if a string qualifies as an fqdn (fully qualified domain name)  */

/* It must contain one or more dots. The dots may not be the first or
	the last character. Two dots may not immidiately follow each other.
	It must consist of the characters a..z, A..Z,, 0..9, '.', '-', '_'.
	The first character must be a letter. */

boolean isfqdn(char *s)
{
	int i=1;
	char c;
	boolean dot=FALSE;

	if (!isalpha(s[0])) return(FALSE);

	while ((c=s[i])!=0)
	{
		if (c=='.')
		{
			dot=TRUE;
			if (s[i-1]=='.') return(FALSE);
		}
		if (!(isalnum(c) || c=='-' || c=='.' || c=='_')) return(FALSE);
		i++;
	}
	if (s[0]=='.' || s[i-1]=='.') return(FALSE);
	return(dot);
}

/* see if a string qualifies as an email address */

/* It must contain exactly one ampersand character that may not be
	the first or the last character. The part after the ampersand must
	contain at least one dot. A dot may not be the first or the last
	character. Two dots may not immidiately follow each other.
	The ampersand may not be immidiately followed or  poreceded by a dot.
	It must consist of the characters a..z, A..Z, 0..9, '.', '-', '_', '@'.
	The first character must be a letter.  */

boolean isemailaddress(char *s)
{
	int i=1, ampersand=0;
	char c;
	boolean dot=FALSE;

	if (!isalpha(s[0])) return(FALSE);

	while ((c=s[i])!=0)
	{
		if (c=='@')
		{
			ampersand++;
			if (s[i-1]=='.') return(FALSE);
		}
		if (c=='.' && i>0)
		{
			if (ampersand) dot=TRUE;
			if (s[i-1]=='.' || s[i-1]=='@') return(FALSE);
		}
		if (!(isalnum(c)||c=='-'||c=='.'||c=='_'||c=='@')) return(FALSE);
		i++;
	}
	if (s[0]=='.' || s[i-1]=='.') return(FALSE);
	if (s[0]=='@' || s[i-1]=='@') return(FALSE);

	return (dot && (ampersand==1));
}

/* See if the flag is of specified type (USER, REGULAR, EMAIL or IP) */

boolean isoftype(char *s, int t)
{
	int i,j=0,l;
	char dest[MAXFLAGLENGTH], *p;

	l=0;
	while (s[l] && s[l]!=':') dest[l]=s[l++];
	dest[l]=0;
	switch(t)
	{
		case USERTYPE		: p=userflags; break;
		case REGULARTYPE	: p=regularflags; break;
		case IPTYPE			: p=ipflags; break;
		case EMAILTYPE		: p=emailflags; break;
	}

	do
	{
		i=0;
		while (p[j+i] && p[j+i]!=',') i++;
		if (i==l && strncmpwild(&p[j],dest,l)==0) return(TRUE);
		j=j+i+1;
	} while (p[j-1]);
	return(FALSE);
}

/* remove a letter from a string */

int removeletter(char *s, int pos)
{
	int i;

	i=pos;
	do
	{
		s[i]=s[i+1];
	}	while (s[i++]);
	return(i-pos);
}

/* Calculate the CRC.
	On entry a character is in ch.
	This is a cumulative process, all chars
	CRC's are taken as a running total.
	To start a new sequence, first clear crc.
	- Inspired by PA0MMV's REMOTE! -  */


unsigned int chcrc(unsigned char ch, unsigned int crc)

{
	unsigned char b,crch;

	b=(crc>>8)^ch;
	crch=(crc&0xFF)^CRC1[b];
	return((crch<<8)+CRC2[b]);
}

/* Update the entrys in the request processors magic list(s) */

void updatemagics(void)
{
	int i=0, j, k, l, magipos, f, alhandle, nrafiles=1;
	int indx[MAXMAGICS+1];
	longstring aliasfile, newfile, actualfile;
	megalongstring mline;
	char magicname[20];
	boolean matc, match[MAXMAGICS];
	FILE *afile, *nfile;

/* First sort out the number of alias files. To do this sort the
	magics and count the number of different words. */

	qsort(magics,nrmagics,sizeof(magics[0]),sort_function);
	indx[0]=0;

/*	'nrafiles' holds the number of different alias files and the array
	'indx' contains the markers for the magics array to indicate an alias
	file change */

	for (i=0;i<nrmagics;i++)
	{
		match[i]=FALSE;

		j=0;
		while (magics[i][j] && magics[i][j]!=' ' && j<sizeof(newfile)-2)
																newfile[j]=magics[i][j++];
		newfile[j]=0;

		indx[nrafiles]=i+1;

		if (i!=0)
		{
			if (strcmp(newfile,aliasfile)!=0) nrafiles++;
		}
		strcpy(aliasfile,newfile);
	}
	for (i=0;i<nrafiles;i++)	/* this is the main loop, done once for */
	{									/* each different alias file */
		j=0;
		while (magics[i][j] && magics[i][j]!=' ' && j<sizeof(newfile)-2)
																newfile[j]=magics[i][j++];
		newfile[j]=0;				/* the alias file name is extracted and */
		if (magics[i][j]) j++;	/* resides now in 'newfile' */
		magipos=j;              /* 'magipos' points to the actual magic name */

		strcpy(newfile,aliasfile);
		force_ext(newfile,".$NW");	/* the new file to write gets a temp EXT */
		alhandle=open(aliasfile,O_RDONLY|O_DENYNONE);
		if (alhandle!=-1) afile=fdopen(alhandle,"rt");
		if (alhandle==-1 || afile==NULL)
		{
			Error(22,0,NULL,aliasfile,NULL,0,NULL,0,0,FALSE);
			break;
		}
		nfile=fopen(newfile,"wt");
		if (nfile==NULL)
		{
			Error(22,0,NULL,newfile,NULL,0,NULL,0,0,FALSE);
			fclose(afile);
			break;
		}
		sprintf(pstr,"Updating %s\n",aliasfile);
		pstring(pstr,0);

		while (Fgets(mline,sizeof(mline)-1,afile)!=NULL)	/* Now step trough */
		{															/* the file line by line */
			matc=FALSE;
			for (j=indx[i];j<indx[i+1];j++)	/* match the magic in this line */
			{											/* against the appropriate magics */
				l=0;                          /* in the array 'magics' */
				while (magics[j][l+magipos] && magics[j][l+magipos]!=' ' &&
						l<sizeof(magicname)-2)	magicname[l]=magics[j][magipos+l++];
				magicname[l]=0;
				if (magics[j][magipos+l]) l++; f=magipos+l;
				k=strlen(magicname);
				if (strnicmp(mline,magicname,k)==0 && isspace(mline[k]))
				{						/* if it matches extract the actual file name */
					matc=match[j]=TRUE;
					substactfile(actualfile,&magics[j][f],FALSE);
					sprintf(pstr,"%s %s\n",magicname,actualfile);
					fprintf(nfile,"%s",pstr);
					pstring(pstr,0);
					break;
				}
			}
			if (!matc) fprintf(nfile,"%s\n",mline);
		}
		for (j=indx[i];j<indx[i+1];j++)
		{
			if (!match[j])		/* handle magics that were not already present */
			{
				substactfile(actualfile,&magics[j][magipos],TRUE);
				sprintf(pstr,"%s\n",actualfile);
				fprintf(nfile,"%s",pstr);
				pstring(pstr,0);
			}
		}
		if (fclose(afile)==0 && fclose(nfile)==0)
		{
			force_ext(newfile,".BAK");
			/* remove an existing .BAK file */
			if (remove(newfile)==-1 && errno!=ENOENT)
			{
				/* if the remove failed for any other reason than non existence */
				Error(11,0,NULL,newfile,NULL,0,NULL,0,0,FALSE);
			}
			else			/* rename old to .BAK en $NW to existing EXT. */
			{
				if (rename(aliasfile,newfile)==-1)
									Error(13,0,NULL,aliasfile,NULL,0,NULL,0,0,FALSE);
				force_ext(newfile,".$NW"); rename(newfile,aliasfile);
			}
		}
	}
}

/* substitute $P$ and $D$ by the actual file name and patch
	the first letter of the extension if a compression parameter
	is supplied */

void substactfile(char *f, char *m, boolean mag)
{
	int j,l,k;
	char magi[20], compresschar=0, *fullname;
	longstring fn;

	l=0;
	if (mag)	while (m[l] && m[l]!=' ') magi[l]=m[l++];
	magi[l]=0;

	if (m[l]==' ') l++;
	k=0;
	while (m[l] && m[l]!=' ') fn[k++]=m[l++];

	fn[k]=0;
	if (m[l]) l++;
	if (stricmp(fn,"$P$")==0) strcpy(fn,pointlist);
	if (stricmp(fn,"$D$")==0) strcpy(fn,diflist);

	/* substitute full name and path if available for the filename */
	if ((fullname=searchpath(fn))!=NULL) strcpy(fn,fullname);

	if (m[l]) compresschar=toupper(m[l]);
	if (compresschar)
	{
		l=max(strlen(fn)-1,0);
		j=0;
		while (fn[l]!='.' && j<=3) {l--; j++;} /* hunt for the dot */
		if (fn[l]=='.')
		{
			if (!fn[l+1]) fn[l+2]=0;
			fn[l+1]=compresschar;      /* replace first char of EXT */
		}
	}
	if (mag)
	{
		strcpy(f,magi);
		strcat(f," ");
		strcat(f,fn);
	}
	else strcpy(f,fn);
}

/* Custum replacements of findfirst and findnext */

int Findfirst(const char *pathname, struct ffblk *ff_blk,int attrib)
{
	int n=-1,result;
	longstring mat;
	char namef[MAXFILE],extf[MAXEXT],fname[13];

	strcpy(mat,pathname);
	while (mat[++n])	if (mat[n]=='#' || mat[n]=='@') mat[n]='?';
	fnsplit(mat,NULL,NULL,namef,extf);
	if ((result=findfirst(mat,ff_blk,attrib))==0)
	{
		fnsplit(pathname,NULL,NULL,namef,extf);
		strcpy(fname,namef); strcat(fname,extf);

		while (result==0 && !supermatch(fname,ff_blk->ff_name))
		{
			result=findnext(ff_blk);
		}
	}
	return(result);
}

int Findnext(char *pathname,struct ffblk *ff_blk)
{
	int result;
	char namef[MAXFILE],extf[MAXEXT],fname[13];

	fnsplit(pathname,NULL,NULL,namef,extf);
	strcpy(fname,namef); strcat(fname,extf);
	do
	{
		result=findnext(ff_blk);
	}	while (result==0 && !supermatch(fname,ff_blk->ff_name));
	return(result);
}

boolean supermatch(char *s, char *d)
{
	int n=0, m=0;

	while (s[n])
	{
		 switch (s[n])
		 {
			case '?':	if (d[m]==0 || d[m]=='.')  return(FALSE);	break;
			case '@':	if (!isalpha(d[m])) return(FALSE); break;
			case '#':	if (!isdigit(d[m])) return(FALSE); break;
			case '*':	while (s[++n]=='*') ;	/* skip multple * */
							if (s[n]=='.')
							{
								while (d[m] && d[m]!='.') m++;
								if (d[m]=='.') continue;
							}
							return(TRUE);
			default :	if (toupper(s[n])!=toupper(d[m])) return(FALSE);
							break;
		 }
		 n++; m++;
	}
	return(TRUE);
}

int editline(char *fn, char *line,char *text,int pos,int maxlen)
{
/*	This subroutine uses direct screen writes. Portability is compromised.
	It can't be helped. It is well neigh impossible to make a screen
	editor without direct screen writes. Results onder another operating
	system than DOS may vary. */

/*	Variables:
	fn		-	Pointer to the input file name
	line	-	Pointer to the line in error
	text	-	Pointer to the text to be edited
	pos	-	Position of the character in text where the error occurs
	maxlen-	Langth of the string to be edited.	*/

	/* window definitions */

	#define WTOP 6
	#define WBOT 10
	#define WLEFT 4
	#define WRIGHT 76

	/* Key definitions */

	#define CR 0x0D
	#define ESC 0x1B
	#define BACKSP 0x08
	#define DEL 21248
	#define LEFT 19200
	#define RIGHT 19712
	#define HOME 18176
	#define END 20224
	#define INS 20992

	#define INSC ((RED<<4)+LIGHTGREEN)
	#define INSCM (LIGHTGRAY)

	int screenbuf[(WBOT-WTOP+1) * (WRIGHT-WLEFT+1)];
	struct text_info r;
	int result=0, c, i, llen, epos, edit=0;
	int insc,backgr,forgr,forgr2,bar,bartxt;
	char templine[WRIGHT-WLEFT-5];

	const int ins[]={'I'+(INSC<<8),'n'+(INSC<<8),'s'+(INSC<<8)};
	const int ovr[]={'O'+(INSC<<8),'v'+(INSC<<8),'r'+(INSC<<8)};
	const int insm[]={'I'+(INSCM<<8),'n'+(INSCM<<8),'s'+(INSCM<<8)};
	const int ovrm[]={'O'+(INSCM<<8),'v'+(INSCM<<8),'r'+(INSCM<<8)};

	if (mono) {insc=INSCM; backgr=BLACK; forgr=WHITE; forgr2=LIGHTGRAY; bar=LIGHTGRAY; bartxt=BLACK;}
	else {insc=INSC; backgr=RED; forgr=WHITE; forgr2=WHITE; bar=GREEN; bartxt=YELLOW; }

	sound(800); delay(200); sound(1000); delay(200); nosound();

	gettextinfo(&r);		/* save the relevant part of the screen */
	gettext(WLEFT,WTOP,WRIGHT,WBOT,screenbuf);

	window(WLEFT,WTOP,WRIGHT,WBOT);		/* Build the edit window */
	textcolor(forgr); textbackground(backgr);
	clrscr();
	gotoxy(2,1);
	cprintf("");
	for (i=3;i<WRIGHT-WLEFT;i++) cprintf("");
	cprintf("");
	for (i=2;i<=WBOT-WTOP;i++)
	{
		gotoxy(2,i);	cprintf("");
		gotoxy(WRIGHT-WLEFT,i); cprintf("");
	}
	gotoxy(2,WBOT-WTOP+1);
	cprintf("");
	for (i=3;i<WRIGHT-WLEFT;i++) cprintf("");
	cprintf("");
	gotoxy(5,1);
	i=strlen(fn)-(WRIGHT-WLEFT-7);
	if (i<0) i=0;
	textcolor(insc&0xFF);
	cprintf(" %s ",&fn[i]);
	gotoxy(WRIGHT-11,WBOT-WTOP+1);
	cprintf("     ");
	puttext(WRIGHT-7,WBOT,WRIGHT+(sizeof(ins)/sizeof(int))-7-1,WBOT,
									 mono ? insert ? insm:ovrm : insert ? ins:ovr);
	window(WLEFT+2,WTOP+1,WRIGHT-2,WBOT-1);
	gotoxy(2,1);
	textcolor(forgr2);
	cprintf("%s",text);
	gotoxy(2,WBOT-WTOP-1);
	cprintf("Edit line; press ENTER to accept, ESC to skip");
	window(WLEFT+3,(WTOP+WBOT)/2,WRIGHT-3,(WTOP+WBOT)/2);
	textbackground(bar);
	textcolor(bartxt);
	clrscr();
	strncpy(templine,line,sizeof(templine)-1);
	templine[sizeof(templine)-1]=0;
	cprintf("%s",templine);
	epos=max(0,pos);
	gotoxy(epos+1,1);

	/*	The main editor loop */

	while (((c=bioskey(0))&0xFF)!=ESC && (c&0xFF)!=CR)
	{
		if (c&0xFF) c&=0xFF;
		llen=strlen(line);
		if (isprint(c))		/* if it is a printable char */
		{
			if (insert && epos<(WRIGHT-WLEFT-6))
			{
				if (llen<maxlen)
				{
					for (i=llen;i>=epos;i--) line[i+1]=line[i];
					line[epos]=c;
					strncpy(templine,line,sizeof(templine)-1);
					templine[sizeof(templine)-1]=0;
					cprintf("%s",&templine[epos]);
				}
			}
			else
			{
				cprintf("%c",c);
				if (line[epos]==0) line[epos+1]=0;
				line[epos]=c;
			}
			if (epos<WRIGHT-WLEFT-6 && epos<maxlen) epos++;
		}
		else
		{
			switch(c)	/* if it is a control character */
			{
				case	LEFT	:	if (epos>0) epos--; break;
				case	RIGHT	:	if (epos<WRIGHT-WLEFT-6 && epos<llen)
									epos++;	break;
				case	HOME	:	epos=0; break;
				case	END	:	epos=llen; break;
				case	INS	:	insert=!insert;
									puttext(WRIGHT-7,WBOT,WRIGHT+(sizeof(ins)/
										sizeof(int))-7-1,WBOT,
										mono ? insert ? insm:ovrm : insert ? ins:ovr);
									break;
				case	DEL	:  for (i=epos;i<llen;i++) line[i]=line[i+1];
									llen--;
									strncpy(templine,line,sizeof(templine)-1);
									templine[sizeof(templine)-1]=0;
									cprintf("%s",&templine[epos]); cprintf(" ");
									break;
				case	BACKSP:  if (epos>0)
									{
										epos--;
										gotoxy(epos+1,1);
										for (i=epos;i<llen;i++) line[i]=line[i+1];
										strncpy(templine,line,sizeof(templine)-1);
										templine[sizeof(templine)-1]=0;
										cprintf("%s",&templine[epos]); cprintf(" ");
										llen--;
									}
									break;
				default:			sound(1000); delay(200); nosound();
									break;
			}
		}
		gotoxy(epos+1,1);
		edit=1;
	}
	if ((c&0xFF)==ESC) result=-1; else result=edit;

	/* restore the screen and exit */
	window(r.winleft,r.wintop,r.winright,r.winbottom);
	puttext(WLEFT,WTOP,WRIGHT,WBOT,screenbuf);
	textattr(r.attribute);
	gotoxy(r.curx,r.cury);
	return(result);
}

/* that's it folks */
